# Some settings require custom widgets to be displayed in the GUI. These are defined in
# this module.

# This workaround is for Python < 3.11. It is not needed in Python 3.11 and later.
try:
    from enum import StrEnum
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum):
        """
        A class that allows for string values in an Enum pre-Python 3.11.
        """


import ipywidgets as ipw
import papermill as pm
import traitlets as tr
from astropy.utils.data import get_pkg_data_filename
from ipyautoui.autoobject import AutoObject
from ipyautoui.custom.iterable import ItemBox
from pydantic import ValidationError
from pydantic.alias_generators import to_snake

from stellarphot.io import tess_photometry_setup
from stellarphot.io.tess import TIC_regex
from stellarphot.settings import (
    Camera,
    Observatory,
    PartialPhotometrySettings,
    PassbandMap,
    PhotometryRunSettings,
    PhotometryWorkingDirSettings,
    SavedSettings,
    ui_generator,
)
from stellarphot.settings.fits_opener import FitsOpener

__all__ = ["ChooseOrMakeNew", "Confirm", "SettingWithTitle"]

DEFAULT_BUTTON_WIDTH = "300px"


class ChooseOrMakeNew(ipw.VBox):
    """
    Widget to present a list of existing items or the option to make a new one.

    Parameters
    ----------

    item_type_name : str
        Name of the item type to be displayed in the widget. Must be one of
        "camera", "observatory", "passband_map", "Camera", "Observatory"
        or "PassbandMap".
    """

    _known_types = [
        "camera",
        "observatory",
        "passband_map",
        Camera.__name__,
        Observatory.__name__,
        PassbandMap.__name__,
    ]

    def __init__(self, item_type_name, *arg, details_hideable=False, **kwargs):
        if item_type_name not in self._known_types:
            raise ValueError(
                f"Unknown item type {item_type_name}. Must "
                f"be {', '.join(self._known_types)}"
            )
        # Get the widgety goodness from the parent class
        super().__init__(*arg, **kwargs)

        self._saved_settings = SavedSettings()
        self._item_type_name = item_type_name

        # keep track of whether we are editing an existing item
        self._editing = False

        # also track whether we are in the midst of a delete confirmation
        self._deleting = False

        # and track if we are making a new item
        self._making_new = False

        # keep track of whether there is a "show details" checkbox
        self._show_details_shown = details_hideable

        self._display_name = item_type_name.replace("_", " ")

        # Create the child widgets

        # Descriptive title
        self._title = ipw.HTML(
            value=(f"Choose a {self._display_name} " "or make a new one")
        )

        self._choose_detail_container = ipw.HBox(layout={"width": DEFAULT_BUTTON_WIDTH})

        # Selector for existing items or to make a new one
        self._choose_existing = ipw.Dropdown(description="")
        choose_width = 75  # percent, the details checkbox takes up the rest
        self._choose_existing.layout.width = (
            f"{choose_width}%" if details_hideable else "100%"
        )

        # Option to show/hide details, only displayed if user wants it.
        self._show_details_ui = ipw.Checkbox(description="Details", value=True)
        self._show_details_ui.layout.display = "flex" if details_hideable else "none"
        # Removes unused whitespace before the checkbox
        self._show_details_ui.style.description_width = "0px"

        if details_hideable:
            self._show_details_ui.layout.width = f"{100 - choose_width}%"
        self._show_details_cached_value = self._show_details_ui.value

        self._choose_detail_container.children = [
            self._choose_existing,
            self._show_details_ui,
        ]

        self._edit_delete_container = ipw.HBox(
            # width below was chosen to match the dropdown...would prefer to
            # determine this programmatically but don't know how.
            layout={"width": DEFAULT_BUTTON_WIDTH}
        )

        self._edit_button = ipw.Button(
            description=f"Edit this {self._display_name}",
        )

        self._delete_button = ipw.Button(
            description=f"Delete this {self._display_name}",
        )

        # Put almost everything into a VBox
        self._details_box = ipw.VBox()

        self._edit_delete_container.children = [self._edit_button, self._delete_button]

        self._confirm_edit_delete = Confirm()

        self._item_widget, self._widget_value_new_item = self._make_new_widget()

        # Put all of the details into a box that can be easily hidden
        self._details_box.children = [
            self._edit_delete_container,
            self._confirm_edit_delete,
            self._item_widget,
        ]

        # Build the main widget
        self.children = [
            self._title,
            self._choose_detail_container,
            self._details_box,
        ]

        # Set up the dropdown widget
        self._construct_choices()
        # Set the selection to the first choice if there is one
        self._choose_existing.value = self._choose_existing.options[0][1]

        # An observer has not been set up yet, so manually call the handler
        if len(self._choose_existing.options) == 1:
            # There are no items, so we are making a new one
            self._handle_selection({"new": "none"})
        else:
            self._handle_selection({"new": self._choose_existing.value})

        # A couple of styling choices for the way existing objects appear
        # in this UI. The title/description is clear from the title of this
        # widget.
        self._item_widget.show_title = False
        self._item_widget.show_description = False
        # Really only applies to PassbandMap, which has nested models,
        # but does no harm in the other cases (true of both lines below)
        self._item_widget.open_nested = True

        # Set up some observers

        # Respond to user clicking the edit button
        self._edit_button.on_click(self._edit_button_action)

        # Respond to user clicking the delete button
        self._delete_button.on_click(self._delete_button_action)

        # Respond to user clicking the save button
        self._item_widget.savebuttonbar.fns_onsave_add_action(self._save_confirmation())

        # Respond to user interacting with a confirmation widget
        # Hide the save button bar so the user gets the confirmation instead
        self._confirm_edit_delete.widget_to_hide = self._item_widget.savebuttonbar
        # Add the observer
        self._confirm_edit_delete.observe(self._handle_confirmation(), names="value")

        # Respond when user wants to make a new thing
        self._choose_existing.observe(self._handle_selection, names="value")

        # Set up an observer to show/hide the details box if the check box
        # is clicked
        self._show_details_ui.observe(self._show_details_handler, names="value")

    @property
    def value(self):
        """
        The value of the widget.
        """
        return self._item_widget.model(**self._item_widget.value)

    @property
    def display_details(self):
        """
        Whether the details box is displayed. Returns the value of the details checkbox
        if the details are hideable, otherwise returns None.
        """
        if self._show_details_shown:
            return self._show_details_ui.value
        else:
            return None

    @display_details.setter
    def display_details(self, value):
        """
        Set the value of the details checkbox if the details are hideable.
        """
        if self._show_details_shown:
            self._show_details_ui.value = value

    def _save_confirmation(self):
        """
        Function to attach to the save button to show the confirmation widget if
        the save button was clicked while editing an existing item.
        """

        def f():
            # This function will be run every time the save button is clicked but
            # we only want to ask for confirmation if we are editing an existing item
            # rather than saving a new one.
            if self._editing:
                self._set_confirm_message()
                self._confirm_edit_delete.show()

        return f

    def _construct_choices(self):
        """
        Set up the choices for the selection widget.
        """
        saved_items = self._saved_settings.get_items(self._item_type_name)
        existing_choices = [(k, v) for k, v in saved_items.as_dict.items()]
        existing_choices = sorted(existing_choices, key=lambda x: x[0].lower())
        choices = existing_choices + [(f"Make new {self._display_name}", "none")]
        # self._choose_existing.value = None
        # This sets the options but doesn't select any of them
        self._choose_existing.options = choices

    def _handle_selection(self, change):
        if change["new"] is None:
            return
        if change["new"] == "none":
            # We are making a new item...

            # Hide the edit button
            self._edit_delete_container.layout.display = "none"

            # Make sure details are shown and hide the "show details" checkbox
            if self._show_details_shown:
                self._show_details_cached_value = self._show_details_ui.value
                self._show_details_ui.value = True
                self._show_details_ui.layout.display = "none"

            # This sets the ui back to its original state when created, i.e.
            # everything is empty.
            self._item_widget._init_ui()

            # Fun fact: _init_ui does not reset the value of the widget. Also,
            # setting the value fails if you try to set it to an empty dict because that
            # is not a valid value for the pydantic model for the widget.
            # So we have to set each of the values individually.
            for key, value in self._widget_value_new_item.items():
                self._item_widget.value[key] = value

            self._item_widget.show_savebuttonbar = True
            self._item_widget.disabled = False

            # Set the validation status to invalid since the user must
            # fill in the fields, and display the validation status
            self._item_widget.is_valid.value = False
            self._item_widget.is_valid.layout.display = "flex"

            # Really only applies to PassbandMap, which has nested models,
            # but does no harm in the other cases (true of both lines below)
            # (and yes, both lines below are needed...this is a bug in ipyautoui,
            #  I think, because open_nested=True isn't respected when we _init_ui.
            #  Forcing a *change* in the value triggers the behavior we want.)
            self._item_widget.open_nested = False
            self._item_widget.open_nested = True

            # Note that we are making a new item
            self._making_new = True

        else:
            # Display the selected item...
            self._item_widget.show_savebuttonbar = False
            self._item_widget.disabled = True
            self._item_widget.is_valid.layout.display = "none"
            self._item_widget.value = self._get_item(change["new"].name)

            # Display the edit button
            self._edit_delete_container.layout.display = "flex"

            # Really only applies to PassbandMap, which has nested models,
            # but does no harm in the other cases
            self._set_disable_state_nested_models(self._item_widget, True)

            # We may have arrived here by choosing a different item while
            # making a new one, so we restore the state of the "show details"
            # checkbox.
            if self._show_details_shown:
                self._show_details_ui.layout.display = "flex"
                self._show_details_ui.value = self._show_details_cached_value

    def _edit_button_action(self, _):
        """
        Handle the edit button being clicked.
        """
        # Replace the display of the edit button with the save button bar...
        self._edit_delete_container.layout.display = "none"
        self._item_widget.show_savebuttonbar = True
        # ...enable the widget...
        self._item_widget.disabled = False
        # ...and show the validation status
        self._item_widget.is_valid.layout.display = "flex"

        # Enable the nested model components
        self._set_disable_state_nested_models(self._item_widget, False)

        # disable the name control, since the whole point of this
        # is to be able to replace the values for a particular name
        self._item_widget.di_widgets["name"].disabled = True

        # This really only applies to PassbandMap, which has nested models,
        # but does no harm in the other cases (true of both lines below)
        # (and yes, both lines below are needed...this is a bug in ipyautoui,
        #  I think, because open_nested=True isn't respected when we _init_ui.
        #  Forcing a *change* in the value triggers the behavior we want.)
        self._item_widget.open_nested = False
        self._item_widget.open_nested = True

        # Update the current state of the widget
        self._editing = True

        # Enable the revert button so that the user can cancel the edit
        self._item_widget.savebuttonbar.bn_revert.disabled = False

    def _delete_button_action(self, _):
        """
        Handle the delete button being clicked.
        """
        # Change our state
        self._deleting = True

        # Hide the edit/delete buttons
        self._edit_delete_container.layout.display = "none"

        # Show the confirmation widget
        self._set_confirm_message()
        self._confirm_edit_delete.show()

    def _show_details_handler(self, change):
        """
        Show or hide the details box based on the value of the checkbox.
        """
        if self._show_details_ui.layout.display == "none":
            # The element is hidden, so just return
            return

        self._details_box.layout.display = "flex" if change["new"] else "none"

    def _set_disable_state_nested_models(self, top, value):
        """
        When a one model contains another and the top-level model widget
        sets disabled=True that does not actually disable the nested model.
        This method handles that in a crude way by walking the tree of
        widgets in the top-level model widget and disabling them all.

        Parameters
        ----------

        top : `ipyautoui.AutoUi`
            Top-level widget that may have nested models.

        value : bool
            State that ``disabled`` should be set to.
        """

        if isinstance(top, AutoObject):
            top.disabled = value
        elif isinstance(top, ItemBox):
            if value:
                # Disabled, so do not show the add/remove buttons
                top.add_remove_controls = "none"
            else:
                # Enabled, so show the add/remove buttons
                top.add_remove_controls = "add_remove"

        try:
            for child in top.children:
                self._set_disable_state_nested_models(child, value)
        except AttributeError:
            # No children...
            pass

    def _set_confirm_message(self):
        """
        Set the message for the confirmation widget.
        """
        if self._editing or self._making_new:
            self._confirm_edit_delete.message = (
                f"Replace value of this {self._display_name}?"
            )
        elif self._deleting:
            self._confirm_edit_delete.message = f"Delete this {self._display_name}?"

    def _make_new_widget(self):
        """
        Make a new widget for the item type and set up actions for the save button.

        Also returns the initial value of the widget for resetting the widget value.
        """
        match self._item_type_name:
            case "camera" | Camera.__name__:
                new_widget = ui_generator(Camera)
            case "observatory" | Observatory.__name__:
                new_widget = ui_generator(Observatory)
            case "passband_map" | PassbandMap.__name__:
                new_widget = ui_generator(PassbandMap)

        def saver():
            """
            Tries to save the new item, and if it fails, shows the confirmation widget.
            """
            try:
                self._saved_settings.add_item(new_widget.model(**new_widget.value))
            except ValueError:
                # This will happen in two circumstances if the item already exists:
                # 1. User is editing an existing item
                # 2. User is making a new item with the same name as an existing one
                self._set_confirm_message()
                self._confirm_edit_delete.show()
            else:
                # If saving works, we update the choices and select the new item
                self._making_new = False
                if self._show_details_shown:
                    self._show_details_ui.layout.display = "flex"
                    self._show_details_ui.value = self._show_details_cached_value
                update_choices_and_select_new()

        def update_choices_and_select_new():
            """
            Update the choices after a new item is saved, update the choices
            and select the new item.
            """
            if not (self._editing or self._making_new):
                value_to_select = new_widget.model(**new_widget.value)
                self._construct_choices()
                self._choose_existing.value = value_to_select
                # Make sure the edit button is displayed
                self._edit_delete_container.layout.display = "flex"

        def revert_to_saved_value():
            """
            Revert the widget to the saved value and end editing.

            This should only apply while editing. If you are making a new
            item you can either select a different item (if there are any) or
            you really need to make a new one.
            """
            if self._editing:
                # We have a selection so we need to stop editing...
                self._editing = False

                # ...and trigger the selection handler.
                self._handle_selection({"new": self._choose_existing.value})

        # This is the mechanism for adding callbacks to the save button.
        new_widget.savebuttonbar.fns_onsave_add_action(saver)
        new_widget.savebuttonbar.fns_onsave_add_action(update_choices_and_select_new)
        new_widget.savebuttonbar.fns_onrevert_add_action(revert_to_saved_value)

        return new_widget, new_widget.value.copy()

    def _handle_confirmation(self):
        """
        Handle the confirmation of a save operation.
        """

        # Use a closure here to capture the current state of the widget
        def confirm_handler(change):
            """
            This handles interactions with the confirmation widget, which is displayed
            when the user has done any of these things:

            + tried to save a new item with the same name as an existing one
            + tried to save an existing item they have edited
            + tried to delete an existing item.

            The widget has three possible values: True (yes), False (no), and None

            This widget is called when the widget value changes, which can happen two
            ways:

            1. The user clicks the "yes" or "no" button, in which case the value will
                be True or False, respectively.
            2. This handler sets the value to None after the user has clicked Yes or No.

            The second case is the reason most of the handler is wrapped in an
            if statement.
            """
            was_editing = self._editing
            # value of None means the widget has been reset to not answered
            if change["new"] is not None:
                item = self._item_widget.model(**self._item_widget.value)
                if self._editing or self._making_new:
                    # We are done editing/making new regardless of
                    # the confirmation outcome
                    self._making_new = False
                    self._editing = False
                    if change["new"]:
                        # User has said yes to updating the item, which we do by
                        # deleting the old one and adding the new one.
                        self._saved_settings.delete_item(item, confirm=True)
                        self._saved_settings.add_item(item)
                        # Rebuild the dropdown list
                        self._construct_choices()
                        # Select the edited item
                        # To make 100% sure the observer is triggered, we set the value
                        # to None first.
                        self._choose_existing.value = None
                        self._choose_existing.value = item
                    else:
                        # User has said no to updating the item, so we just
                        # act as though the user has selected this item.
                        if was_editing:
                            # The user has presumably changed the value in the UI, so
                            # get the correct value from disk.
                            item = self._get_item(item.name)

                            # To make 100% sure the observer is triggered, we set the
                            # value to None first.
                            self._choose_existing.value = None
                            self._choose_existing.value = item
                        else:
                            # Set the selection to the first choice if there is one
                            # To make 100% sure the observer is triggered, we set the
                            # value to None first.
                            self._choose_existing.value = None
                            self._choose_existing.value = self._choose_existing.options[
                                0
                            ][1]

                elif self._deleting:
                    if change["new"]:
                        # User has confirmed the deletion
                        self._saved_settings.delete_item(item, confirm=True)
                        # Rebuild the dropdown list
                        self._construct_choices()

                        # Select the first item...
                        self._choose_existing.value = self._choose_existing.options[0][
                            1
                        ]
                        # ...but if there is only one option, the line above doesn't
                        # trigger the _choose_existing observer because the value is set
                        # when the options are set. So we need to trigger it manually.
                        if len(self._choose_existing.options) == 1:
                            self._handle_selection({"new": self._choose_existing.value})
                    else:
                        # User has decided not to delete the item
                        self._handle_selection({"new": item})
                    self._deleting = False

                # Reset the confirmation widget to unanswered
                self._confirm_edit_delete.value = None

        return confirm_handler

    def _get_item(self, item_name):
        """
        Get an item from the saved settings by name.
        """
        match self._item_type_name:
            case "camera" | Camera.__name__:
                container = self._saved_settings.cameras
            case "observatory" | Observatory.__name__:
                container = self._saved_settings.observatories
            case "passband_map" | PassbandMap.__name__:
                container = self._saved_settings.passband_maps

        return container.as_dict[item_name]


class Confirm(ipw.HBox):
    """
    Widget to confirm a choice.

    The value of this widget will be ``True`` if the user confirms the choice, ``False``
    if they do not, and ``None`` if they have not yet answered.
    """

    def __init__(self, message="", widget_to_hide=None, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        # Hide this widget until it is needed
        self.layout.display = "none"
        self._other_widget = widget_to_hide
        self._message = ipw.HTML(value=message)
        button_layout = ipw.Layout(width="50px")
        self._yes = ipw.Button(
            description="Yes", button_style="success", layout=button_layout
        )
        self._no = ipw.Button(
            description="No", button_style="danger", layout=button_layout
        )
        self._yes.on_click(self._handle_yes)
        self._no.on_click(self._handle_no)
        self.children = [self._message, self._yes, self._no]
        # Value van be either True (yes), False (no), or None (not yet answered)
        self.add_traits(value=tr.Bool(allow_none=True))
        self.value = None

    @property
    def message(self):
        return self._message.value

    @message.setter
    def message(self, value):
        self._message.value = value

    def show(self):
        """
        Display the confirmation widget and, if desired, hide the widget it replaces.
        """
        self.layout.display = "flex"
        if self._other_widget is not None:
            self._other_widget.layout.display = "none"

    # THere ought to be a way to refactor these two, but this works for now.
    def _handle_yes(self, _):
        self.layout.display = "none"
        if self._other_widget is not None:
            self._other_widget.layout.display = "flex"
        self.value = True

    def _handle_no(self, _):
        self.layout.display = "none"
        if self._other_widget is not None:
            self._other_widget.layout.display = "flex"
        self.value = False


class SaveStatus(StrEnum):
    """
    Class to define the symbols used to represent a save status.
    """

    SETTING_NOT_SAVED = "❗️"
    SETTING_IS_SAVED = "✅"
    SETTING_SHOULD_BE_REVIEWED = "🔆"


class SettingWithTitle(ipw.VBox):
    """
    Class that adds a title to a setting widget made by ipyautoui and
    styles the title based on whether the settings need to be saved.

    Parameters
    ----------

    plain_title : str
        Title of the setting widget without any decoration.

    widget : ipyautoui.AutoUi
        The setting widget to be displayed.
    """

    badge = tr.UseEnum(SaveStatus, default=None, allow_none=True)

    def __init__(self, plain_title, widget, header_level=2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._header_level = header_level
        self._plain_title = plain_title
        self._widget = widget

        if isinstance(widget, ChooseOrMakeNew):
            self.title = self._widget._title
            self.children = [self._widget]
            observer = self._choose_existing_observer
            self._widget._choose_existing.observe(observer, names="value")
            self._autoui_widget = self._widget._item_widget
            # In case a value gets set programmatically....
            # self._autoui_widget.observe(self._title_observer, names="_value")
        else:
            self.title = ipw.HTML()
            self._format_title(None)
            self.children = [self.title, self._widget]
            # Set up an observer to update title decoration when the settings
            # change.
            observer = self._title_observer
            # Also update after the save button is clicked
            # self._widget.savebuttonbar.fns_onsave_add_action(self._title_observer)
            self._autoui_widget = self._widget
        self._autoui_widget.savebuttonbar.observe(observer, names="unsaved_changes")

    def _choose_existing_observer(self, _=None):
        """
        Observer for the ChooseOrMakeNew widget.
        """
        # Unless we are making a new item or editing an item then what is displayed
        # is saved.
        if not self._widget._making_new and not self._widget._editing:
            self.badge = SaveStatus.SETTING_IS_SAVED
        else:
            self.badge = SaveStatus.SETTING_NOT_SAVED

    @tr.observe("badge")
    def _format_title(self, _=None):
        badge = self.badge or ""
        badge = badge + " " if badge else ""
        self.title.value = (
            f"<h{self._header_level}>{badge}{self._plain_title}</h{self._header_level}>"
        )

    def decorate_title(self):
        """
        Public interface for forcing a title update.
        """
        self._format_title()

    def _title_observer(self, change):
        """
        Observer for the title of the widget, triggered when unsaved_changes
        changes.
        """
        if change["new"]:
            # i.e. unsaved_changes is True
            self.badge = SaveStatus.SETTING_NOT_SAVED
        else:
            self.badge = SaveStatus.SETTING_IS_SAVED


class ReviewSettings(ipw.VBox):
    """
    Widget to preview the saved settings in the working directory. It displays one
    tab or accordion for each type of setting being reviewed.

    This widget does a bunch of automatic saving and loading behind the scenes:

    1. When the widget is created, it loads the settings from the working directory, if
       there are any. Settings loaded this way are marked as "need review" to remind the
       user they might want to take a look.
    2. When the widget is created, any of the saveable settings are set to the default
       for that setting and then saved to the working directory, with the tab markked as
       "needs review".
    2. When the user clicks the save button for a setting that displays one,
       the settings are saved to the working directory settings.
    3. When the user selects a setting from the settings that have a dropdown, the
       selected setting is saved. Currently those settings are Camera, Observatory, and
       PassbandMap, but the definitive list is given in
       `stellarphot.settings.ChooseOrMakeNew._known_types`.
    4. Creating a new one of those saveable settings also saves it to the working
       directory settings.

    """

    def __init__(self, settings, style="tabs", *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Get a copy of whatever settings may have already been saved.
        try:
            self._current_settings = PhotometryWorkingDirSettings().load()
        except ValueError:
            self._current_settings = PartialPhotometrySettings()

        self._setting_widgets = []
        self._plain_names = []
        self.badges = []

        self._settings = settings

        for setting in settings:
            # Track whether we are using the ChooseOrMakeNew or not
            is_choose_or_make_new = False
            if setting.__name__ in ChooseOrMakeNew._known_types:
                widget = ChooseOrMakeNew(setting.__name__)
                val_to_set = widget._choose_existing
                is_choose_or_make_new = True
            else:
                widget = ui_generator(setting)
                val_to_set = widget

            _add_saving_to_widget(widget)
            name = to_snake(setting.__name__)
            plain_name = " ".join(name.split("_"))
            self._plain_names.append(plain_name)
            self._setting_widgets.append(SettingWithTitle(plain_name, widget))

            # This should be either a valid object or None
            saved_value = getattr(self._current_settings, name)

            if saved_value is not None:
                try:
                    if is_choose_or_make_new:
                        # Set to None first to ensure there is a change in the value
                        # when we set it to saved_value.
                        val_to_set.value = None
                    val_to_set.value = saved_value
                except tr.TraitError as e:
                    # It can happen, while testing, that a setting gets saved to a local
                    # directory but is no longer in the saved settings for Camera, etc.
                    # We cannot fix that here, so raise a clearer error.
                    raise ValueError(
                        f"The {name} setting saved in the working directory is not "
                        f"consistent with the list of {name} items that are saved "
                        "in your permanent settings. Please fix this manually "
                        f"by editing your saved {name} settings or by deleting the "
                        "working directory settings."
                    ) from e
                # Add symbol to title to indicate that the setting needs review
                self.badges.append(SaveStatus.SETTING_SHOULD_BE_REVIEWED)

            elif is_choose_or_make_new:
                if len(widget._choose_existing.options) > 1:
                    # There is also one already-saved choice, so we load it.
                    # If we are using the ChooseOrMakeNew widget, we need to set the
                    # value of the widget to the default item to trigger the save.
                    # To do that, first set the value to None and then set the value
                    # back to the default item.
                    default_item = val_to_set.value
                    val_to_set.value = None
                    val_to_set.value = default_item
                    self.badges.append(SaveStatus.SETTING_SHOULD_BE_REVIEWED)
                else:
                    # This setting needs to be made, not reviewed
                    self.badges.append(SaveStatus.SETTING_NOT_SAVED)
            else:
                # We got here because there was not a setting saved in the working
                # directory, and this is not a ChooseOrMakeNew, which might have
                # settings saved at the user level.

                # Two possibilities:
                # 1. The setting can be made from default values but needs to be
                #    reviewed. Status should be "needs review".
                # 2. The setting cannot be made from default values. Status should be
                #    "not saved".
                try:
                    val_to_set.model()
                except ValidationError:  # pragma: no cover
                    # This should never happen with the code base as of 2024-06-27 but
                    # might in the future.
                    self.badges.append(SaveStatus.SETTING_NOT_SAVED)
                else:
                    # This setting can be made from default values, so we save it to the
                    # working directory.
                    val_to_set.savebuttonbar.bn_save.click()
                    self.badges.append(SaveStatus.SETTING_SHOULD_BE_REVIEWED)

        # Check that everything is consistent....
        assert len(self.badges) == len(self._plain_names)

        if style == "tabs":
            self._container = ipw.Tab()
        else:
            self._container = ipw.Accordion()

        self._container.children = self._setting_widgets
        self._container.titles = self._make_titles()

        self.children = [self._container]

        # Set up an observer to run when a tab is selected
        self._container.observe(self._observe_tab_selection, names="selected_index")

        # Set up observer for each of the widget badges
        for idx, widget in enumerate(self._setting_widgets):
            widget.observe(self._observe_badge_change(idx), names="badge")

    def _make_titles(self):
        """
        Make titles from badges and plain titles.
        """
        return [
            f"{badge} {plain}"
            for badge, plain in zip(self.badges, self._plain_names, strict=True)
        ]

    @property
    def current_settings(self):
        """
        The current settings in the widget.
        """
        try:
            self._current_settings = PhotometryWorkingDirSettings().load()
        except ValueError:
            self._current_settings = PartialPhotometrySettings()

        return self._current_settings

    def _observe_tab_selection(self, change):
        """
        Observer for the tab or accordion selection.
        """
        # Once the user has clicked on the tab, the status badge for the
        # tab should just be the badge for the widget it holds, if the
        # widget has a badge. Otherwise, compare the widget value to the
        # saved value to determine the badge.

        # Get the index
        new_selected = change["new"]

        setting_badge = self._setting_widgets[new_selected].badge
        if setting_badge is not None:
            self.badges[new_selected] = setting_badge
        else:
            # Check whether the setting is saved or not
            setting_widget = self._setting_widgets[new_selected]

            snake_name = to_snake(setting_widget._autoui_widget.model.__name__)

            disk_value = getattr(self.current_settings, snake_name)
            if disk_value is None:
                # The setting is not saved
                setting_widget = SaveStatus.SETTING_NOT_SAVED
            else:
                # Set the badge to saved if it has been saved.
                value_from_widget = setting_widget._autoui_widget.model.model_validate(
                    setting_widget._autoui_widget.value
                )

                if disk_value == value_from_widget:
                    setting_widget.badge = SaveStatus.SETTING_IS_SAVED

        self._container.titles = self._make_titles()

    def _observe_badge_change(self, index):
        """
        Observer for the badge of a setting widget.
        """

        def observer(change):
            self.badges[index] = change["new"]

            # This should only be called when a badge changes, so we can just
            # update the titles.
            self._container.titles = self._make_titles()

        return observer


def _add_saving_to_widget(setting_widget):
    """
    Add an observer to a widget that autosaves the settings for that widget to
    the working directory.

    Parameters
    ----------
    setting_widget : ChooseOrMakeNew
        The widget to add the observer to.
    """
    wd_settings = PhotometryWorkingDirSettings()

    # Define name here so that it is available in the save_wd function. Its
    # value will be set in the if/elif block below.
    name = ""

    def save_wd(_=None):
        try:
            pps = PartialPhotometrySettings(**{name: setting_widget.value})
        except ValidationError:
            # This can happen while making a new item, or while in the process of
            # editing one
            return
        # We have a validated setting so save it.
        wd_settings.save(pps, update=True)

    if hasattr(setting_widget, "_choose_existing"):
        setting_widget._choose_existing.observe(save_wd, "value")
        name = to_snake(setting_widget._item_type_name)
    elif hasattr(setting_widget, "savebuttonbar"):
        setting_widget.savebuttonbar.fns_onsave_add_action(save_wd)
        name = to_snake(setting_widget.model.__name__)
    else:
        raise ValueError(
            f"The widget {setting_widget} is not a recognized type of widget."
        )


class PhotometryRunner(ipw.VBox):
    """
    Class to run the photometry notebook on a folder of images.

    Parameters
    ----------
    photometry_notebook_name : str, optional
        Name of the photometry notebook to run. Default is "photometry_run.ipynb".

    Notes
    -----

    When this widget is run, it will created a new notebook in the current directory
    called ``photometry_notebook_name``, which will perform the photometry.
    """

    def __init__(
        self, photometry_notebook_name="photometry_run.ipynb", *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.photometry_notebook_name = photometry_notebook_name
        self.fitsopen = FitsOpener(
            title=(
                "Choose an image of the object of interest to you.</br>Photometry "
                "will be performed on all images of that object in the folder "
                "containing the selected image."
            )
        )
        self.info_box = ipw.HTML()
        self.run_output = ipw.Output()
        self.confirm = Confirm(message="Is this correct?")
        self.children = (
            self.fitsopen.file_chooser,
            self.info_box,
            self.confirm,
            self.run_output,
        )
        self.fitsopen.file_chooser.observe(self._file_chosen, "_value")
        self.confirm.observe(self._confirmation, "value")
        self.run_settings = None

    def _file_chosen(self, _):
        self.run_settings = PhotometryRunSettings(
            directory_with_images=self.fitsopen.path.parent,
            object_of_interest=self.fitsopen.header["object"],
        )
        self.info_box.value = (
            "<h2>" + self.info_message + "</br>Is this correct?" + "</h2>"
        )
        self.confirm.show()

    @property
    def info_message(self):
        return (
            f"Photometry will be performed on all images of the object "
            f"'<code>{self.run_settings.object_of_interest}</code>' in the "
            f"folder '<code>{self.run_settings.directory_with_images}</code>'"
        )

    def _confirmation(self, change=None):
        if change["new"]:
            # User said yes

            # Update informational message
            self.info_box.value = (
                "<h2>" + self.info_message + "</br>Photometry is running..." + "</h2>"
            )
            template_nb = get_pkg_data_filename(
                "photometry_runner.ipynb", package="stellarphot.notebooks"
            )

            with self.run_output:
                pm.execute_notebook(
                    template_nb,
                    self.photometry_notebook_name,
                    parameters=self.run_settings.model_dump(mode="json"),
                )
        else:
            # User said no, so reset to initial state.
            self.fitsopen.file_chooser.reset()
            self.info_box.value = ""
            self.run_settings = None


class TessPhotometrySetup(ipw.VBox):
    """
    Widget for getting some photometry inputs for a TESS Object of Interest (TOI).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Define the children of self
        self.header = ipw.HTML("<h2>Enter the TIC ID of the exoplanet candidate</h2>")
        self.drop_label = ipw.HTML("How would you like to specify the TIC ID?")
        self.drop = ipw.Dropdown(
            options=(("Type in TIC ID", 0), ("Choose an image", 1))
        )
        self.tic_id_entry = ipw.IntText(description="TIC ID")
        self.fits_opener = FitsOpener()
        self.confirm = Confirm()
        self.spinner = Spinner(message="<h3>Downloading TIC info...</h3>")
        self.all_done = ipw.HTML("<h2>Done! Files have been written.</h2>")

        # Set up observers
        self.drop.observe(self.watch_drop(), names="value")
        self.tic_id_entry.observe(self.watch_tic_id_text_box(), names="value")
        self.fits_opener.file_chooser.observe(self.watch_fits_opener(), names="_value")
        self.confirm.observe(self.watch_confirmation(), names="value")

        # Initialize the widget

        # Hide the "all done" message until it is needed
        self.all_done.layout.display = "none"

        # Initialize the TIC ID
        self.tic_id = 0

        # This is correct -- watch_drop returns a function that takes a change dict
        # Option 0 is to type in the TIC ID
        self.watch_drop()({"new": 0})

        # Define the children
        self.children = (
            self.header,
            self.drop_label,
            self.drop,
            self.tic_id_entry,
            self.fits_opener.file_chooser,
            self.confirm,
            self.spinner,
            self.all_done,
        )

    def watch_drop(self):
        def observer(change):
            # Change whether a text box for entering the TIC ID is visible or a
            # file chooser for selecting an image is visible.
            if change["new"] == 0:
                self.tic_id_entry.layout.display = "flex"
                self.fits_opener.file_chooser.layout.display = "none"
            else:
                self.tic_id_entry.layout.display = "none"
                self.fits_opener.file_chooser.layout.display = "flex"

        return observer

    def set_tic_id(self, an_id):
        # The actual setting of the TIC ID is here so that all observers can
        # call the same method.
        self.tic_id = int(an_id)
        # Display the confirmation widget
        self.confirm.message = f"Is the TIC ID {self.tic_id} correct?"
        self.confirm.show()

    def watch_tic_id_text_box(self):
        def observer(change):
            # Just set the TIC ID....
            self.set_tic_id(change["new"])

        return observer

    def watch_fits_opener(self):
        def observer(_):
            # The FitsOpener handles loading the header and setting the object.
            # Here we just need to extract the TIC ID from the object.
            match = TIC_regex.match(self.fits_opener.object)
            self.set_tic_id(match.group("star"))

        return observer

    def watch_confirmation(self):
        def observer(change):
            # Confirm has a bool value, True if the user says yes, False if no.
            # The Confirm widget closes itself, so we don't need to do anything
            # here except to start the spinner, run the setup function, and stop
            # the spinner.
            if change["new"]:
                self.spinner.start()
                tess_photometry_setup(self.tic_id, overwrite=True)
                self.spinner.stop()
                self.all_done.layout.display = "flex"

        return observer


class Spinner(ipw.VBox):
    """
    A spinner widget.
    """

    def __init__(self, *args, spinner_file=None, message="", **kwargs):
        if spinner_file is None:
            spinner_file = get_pkg_data_filename(
                "data/star_spinner.svg", package="stellarphot"
            )
        self._spinner_file = spinner_file
        super().__init__(*args, **kwargs)
        with open(spinner_file) as f:
            self._spinner = ipw.HTML(f.read())
        self._message = ipw.HTML(message)
        self.children = [self._message, self._spinner]
        self.layout.display = "none"
        self.layout.width = "200px"

    @property
    def message(self):
        return self._message.value

    @property
    def spinner_file(self):
        return self._spinner_file

    def start(self):
        self.layout.display = "flex"

    def stop(self):
        self.layout.display = "none"



from .models import *
from .settings_files import *
from .views import *

# For conveniences, provide the JSON schema and an example of PhotometrySettings.
photometry_settings_schema = PhotometrySettings.model_json_schema()  # noqa: F405

from .tests.test_models import TEST_PHOTOMETRY_SETTINGS

photometry_settings_example = PhotometrySettings(  # noqa: F405
    **TEST_PHOTOMETRY_SETTINGS
)

# Clean up the namespace
del TEST_PHOTOMETRY_SETTINGS



import re
from pathlib import Path
from typing import ClassVar

from platformdirs import PlatformDirs
from pydantic import BaseModel, ValidationError

from .models import (
    Camera,
    Observatory,
    PartialPhotometrySettings,
    PassbandMap,
    PhotometrySettings,
)

__all__ = ["SavedSettings", "SETTINGS_FILE_VERSION", "PhotometryWorkingDirSettings"]

# We will have to version settings formats, I think. Hopefully this changes rarely
# or never.
SETTINGS_FILE_VERSION = "2"  # value chosen to match major version of stellarphot

ENCODING = "utf-8"


class SavedFileOperations:
    # Provide a place to store the path to the settings file. Annotate as a ClassVar
    # so that pydantic doesn't think it is a field. Also mark it as private to
    # discourage direct access.
    _settings_path: ClassVar = None

    def save(self):
        file_path = self._settings_path / self._file_name
        json_data = self.model_dump_json(indent=4)
        with file_path.open("w", encoding=ENCODING) as f:
            f.write(json_data)

    def get(self, name):
        """
        Get the item with the given name.

        Parameters
        ----------
        name : str
            Name of the item to get.
        """
        return self.as_dict[name]

    @classmethod
    def load_model(cls):
        file_path = cls._settings_path / cls._file_name
        if not file_path.exists():
            instance = cls(as_dict={})
        else:
            with file_path.open(encoding=ENCODING) as f:
                instance = cls.model_validate_json(f.read())

        return instance

    @classmethod
    def delete(cls, confirm=False, name=None):
        """
        Delete the settings file for this class.

        Parameters
        ----------
        confirm : bool, optional
            If True, the file is deleted. If False, a ValueError is raised.

        name : str, optional
            Name of the item to delete. If provided, only the item with this name is
            deleted. If not provided, the entire file is deleted.
        """
        if not confirm:
            raise ValueError("You must confirm deletion by passing confirm=True")

        file_path = cls._settings_path / cls._file_name
        if name is not None:
            # Only delete the named item
            instance = cls.load_model()
            if name not in instance.as_dict:
                raise ValueError(f"{name} not found in {cls._file_name}")
            del instance.as_dict[name]
            instance.save()
        else:
            # Delete the entire file
            file_path.unlink(missing_ok=True)


class Cameras(SavedFileOperations, BaseModel):
    # Using the ClassVar annotation means this is treated as a class variable rather
    # than a pydantic field. We don't want pydantic storing the name of the settings
    # file in the settings file itself.
    _file_name: ClassVar[str] = "cameras.json"
    "Name of the file where the cameras are saved."

    as_dict: dict[str, Camera]
    "Dictionary of cameras, keyed by camera name."


class Observatories(SavedFileOperations, BaseModel):
    _file_name: ClassVar[str] = "observatories.json"
    "Name of the file where the observatories are saved."
    as_dict: dict[str, Observatory]
    "Dictionary of observatories, keyed by observatory name."


class PassbandMaps(SavedFileOperations, BaseModel):
    _file_name: ClassVar[str] = "passband_maps.json"
    "Name of the file where the passband maps are saved."
    as_dict: dict[str, PassbandMap]
    "Dictionary of passband maps, keyed by passband map name."


class SavedSettings:
    """
    Handle loading and saving of settings files from disk.
    """

    def __init__(self, _testing_path=None, _create_path=True):
        """
        Parameters
        ----------

        _testing_path : Path, optional
            Path to use for testing purposes. If not provided, the default path is used.

        _create_path : bool, optional
            If True, the directory where settings files are stored is created if it does
            not exist. If False, the directory is not created, which is useful for
            testing.
        """
        data_dir = PlatformDirs(
            "stellarphot", version=SETTINGS_FILE_VERSION
        ).user_data_dir
        self._settings_path = Path(data_dir)
        if _create_path:
            if not self.settings_path.exists():
                self.settings_path.mkdir(parents=True)

        # Make the path available to the SavedFileOperations classes.
        SavedFileOperations._settings_path = self.settings_path

    @property
    def settings_path(self):
        """
        Path to the directory where settings files are stored.
        """
        return self._settings_path

    @property
    def cameras(self) -> Cameras:
        """
        Cameras stored in the settings.
        """
        # Note that we always reload in case the file has changed.
        return Cameras.load_model()

    @property
    def observatories(self) -> Observatories:
        """
        Observatories stored in the settings.
        """
        return Observatories.load_model()

    @property
    def passband_maps(self) -> PassbandMaps:
        """
        Passband maps stored in the settings.
        """
        return PassbandMaps.load_model()

    def get_items(self, item_type):
        """
        Get the items of a given type.

        Parameters
        ----------
        item_type : str | Camera | Observatory | PassbandMap
            The type of item to get.
        """
        match item_type:
            case Camera() | "camera" | Camera.__name__:
                return self.cameras
            case Observatory() | "observatory" | Observatory.__name__:
                return self.observatories
            case PassbandMap() | "passband_map" | PassbandMap.__name__:
                return self.passband_maps
            case _:
                raise ValueError(
                    f"Unknown item {item_type} of type {type(item_type)}. Must be "
                    "Camera, Observatory, or PassbandMap, or "
                    "'camera', 'observatory', or 'passband_map'"
                )

    def add_item(self, item):
        """
        Add an item to the settings.

        Parameters
        ----------
        item : Camera | Observatory | PassbandMap
            The item to add.
        """
        match item:
            case Camera() as to_add:
                container = self.cameras
            case Observatory() as to_add:
                container = self.observatories
            case PassbandMap() as to_add:
                container = self.passband_maps
            case _:
                raise ValueError(
                    f"Unknown item {item} of type {type(item)}. Must be Camera, "
                    "Observatory, or PassbandMap"
                )

        if to_add.name in container.as_dict:
            raise ValueError(f"{to_add.name} already exists in {container._file_name}")

        container.as_dict[to_add.name] = to_add
        container.save()

    def delete(self, confirm=False, delete_settings_folder=False):
        """
        Delete all settings files.

        Parameters
        ----------
        confirm : bool, optional
            If True, the files are deleted. If False, a ValueError is raised.

        delete_settings_folder : bool, optional
            If True, the directory where settings files are stored is deleted. If False,
            only the settings files are deleted.
        """
        if not confirm:
            raise ValueError("You must confirm deletion by passing confirm=True")
        Cameras.delete(confirm=confirm)
        Observatories.delete(confirm=confirm)
        PassbandMaps.delete(confirm=confirm)
        if delete_settings_folder:
            self.settings_path.rmdir()

    def delete_item(self, item, confirm=False):
        """
        Delete an item from the settings.

        Parameters
        ----------
        item : Camera | Observatory | PassbandMap
            The item to delete.

        confirm : bool, optional
            If True, the item is deleted. If False, a ValueError is raised.
        """
        match item:
            case Camera() as to_delete:
                klass = Cameras
            case Observatory() as to_delete:
                klass = Observatories
            case PassbandMap() as to_delete:
                klass = PassbandMaps
            case _:
                raise ValueError(
                    f"Unknown item {item} of type {type(item)}. Must be Camera, "
                    "Observatory, or PassbandMap"
                )

        klass.delete(confirm=confirm, name=to_delete.name)


class PhotometryWorkingDirSettings:
    """
    Class to save in-progress and complete photometry settings in the working directory.
    """

    def __init__(self, settings_file_name="photometry_settings.json"):
        """
        Parameters
        ----------
        settings_file_name : str, optional
            Name of the settings file. Must end with '.json', contain only
            alphanumeric characters, hyphens, underscores, and spaces and begin
            with an alphanumeric character.
        """
        self._working_dir = Path(".")
        self._check_bad_file_name(settings_file_name)
        self._settings_file = self._working_dir / Path(settings_file_name)
        self._partial_settings_file = self._working_dir / Path(
            "partial_" + settings_file_name
        )
        self._partial_settings = None
        self._settings = None

    @property
    def settings(self):
        """
        The full settings, or None
        """
        return self._settings

    @property
    def partial_settings(self):
        """
        The partial settings, or None
        """
        return self._partial_settings

    # Properties for settings file and partial settings file
    @property
    def settings_file(self):
        return self._settings_file

    @property
    def partial_settings_file(self):
        return self._partial_settings_file

    def _check_bad_file_name(self, file_name):
        good_name = re.compile(r"^\w+[\w\d\-_ ]*\.json$")
        if not good_name.match(file_name):
            raise ValueError(
                f"Settings file name {file_name} is not a valid name. The name can "
                "only contain alphanumeric characters, hyphens, underscores, and "
                "spaces, and must end with '.json'"
            )

    def _are_partial_actually_full(self, settings):
        """
        Check if the partial settings are actually full settings.

        Parameters
        ----------
        settings : PartialPhotometrySettings
            The settings to check.
        """
        try:
            PhotometrySettings.model_validate(settings.model_dump())
        except ValidationError:
            return False
        else:
            return True

    def _update_settings_from_partial(self, disk_settings, partial_settings):
        """
        Update the settings on disk, which may be full or partial, with the partial
        settings.

        Parameters
        ----------
        disk_settings : PhotometrySettings or PartialPhotometrySettings
            The settings on disk.

        partial_settings : PartialPhotometrySettings
            The partial settings to update with.

        Returns
        -------
        PhotometrySettings pr PartialPhotometrySettings
            The updated settings. The return type is the same as the type
            of disk_settings.
        """
        # Grab a dict of the settings, only keeping values that are not
        # None. Making it dict so that the update method can be used to
        # merge the two sets of settings.

        passed_partial_settings = {
            k: v for k, v in partial_settings.model_dump().items() if v is not None
        }

        # The order matters here. Keys in the argument passed_partial_settings
        # will overwrite the keys in disk settings.
        # Note that update works in-place.
        disk_settings.update(passed_partial_settings)

        return disk_settings

    def save(self, settings, update=False):
        """
        Save the partial or full settings to the working directory. Note well that
        this removes any partial settings file if called with valid full settings.

        Parameters
        ----------
        settings : PhotometrySettings
            The settings to save.

        update : bool, optional
            If True, the settings are updated -- see Note below for more description.
            If False, the settings are overwritten.

        Notes
        -----

        If ``update`` is True, then the settings are updated. This means that if the
        settings passed in are partial and there is already a partial setting saved on
        disk, then the settings from disk that are not in the new settings are added to
        the new settings.

        This also means that in the event that the settings in the argument have, say,
        a `~stellarphot.settings.Camera`, and the file on disk also has one, the one
        in the argument is the one that will be kept.

        Finally, if we are passed a partial setting and there is a full setting on disk,
        then the partial settings are merged with the full settings, and the full
        settings are saved.
        """
        full_settings = False

        try:
            _ = self.load()
        except ValueError:
            # If we catch a ValueError, then we are in a situation where we have no
            # settings files. We can proceed to save the settings.
            pass

        match settings:
            case PartialPhotometrySettings():
                # This case MUST come first, because PartialPhotometrySettings is a
                # subclass of PhotometrySettings.

                # Are there already full settings?
                if self._settings_file.exists():
                    if not update:
                        # If so, we can't save partial settings if the update flag
                        # is False.
                        raise ValueError(
                            "Cannot save partial settings when full "
                            "settings already exist."
                        )
                    else:
                        # Load the full settings and update them with the
                        # partial settings
                        disk_settings = self._settings.model_dump()

                        disk_settings = self._update_settings_from_partial(
                            disk_settings, settings
                        )

                        disk_settings = PhotometrySettings.model_validate(disk_settings)

                        settings = disk_settings

                # Are we updating or replacing partial settings?
                if update and self._partial_settings is not None:
                    # Get the partial settings that were loaded from disk
                    existing_partial_settings = self._partial_settings.model_dump()

                    # Update the partial settings with the new settings
                    existing_partial_settings = self._update_settings_from_partial(
                        existing_partial_settings, settings
                    )

                    # Validate the updated partial settings
                    settings = PartialPhotometrySettings.model_validate(
                        existing_partial_settings
                    )
                # set variable file to point to appropriate (partial or full)
                # settings file location.
                if self._are_partial_actually_full(settings):
                    self._settings = settings
                    file = self._settings_file
                    full_settings = True
                else:
                    # Update the partial settings with the new settings
                    self._partial_settings = settings
                    file = self._partial_settings_file
            case PhotometrySettings():
                self._settings = settings
                file = self._settings_file
                full_settings = True

            case _:
                raise ValueError(
                    "Settings must be PhotometrySettings or PartialPhotometrySettings, "
                    f"not {type(settings)}"
                )

        if full_settings:
            self._partial_settings_file.unlink(missing_ok=True)
            self._partial_settings = None

        # Write the settings to a file. The settings themselves are models, so we
        # are guaranteed to write the correct model type (partial or full settings)
        # to the file.
        with file.open("w", encoding=ENCODING) as f:
            f.write(settings.model_dump_json(indent=4))

    def load(self):
        """
        Load full or partial settings.

        Returns
        -------
        PhotometrySettings | PartialPhotometrySettings | None
            The settings loaded from disk, or None if there are no settings files.
        """
        # Assume we have nothing to begin....
        self._partial_settings = None
        self._settings = None

        if not (self._settings_file.exists() or self._partial_settings_file.exists()):
            raise ValueError(f"Settings file {self._settings_file} does not exist")

        # Load PartialPhotometrySettings first, if it exists.
        if self._partial_settings_file.exists():
            with self._partial_settings_file.open(encoding=ENCODING) as f:
                content = f.read()

            try:
                self._partial_settings = PartialPhotometrySettings.model_validate_json(
                    content
                )
            except ValidationError as e:
                raise ValueError(f"Error loading partial settings: {e}") from e

        # Now load full settings if they exist
        if self._settings_file.exists():
            with self._settings_file.open(encoding=ENCODING) as f:
                content = f.read()

            try:
                self._settings = PhotometrySettings.model_validate_json(content)
            except ValidationError as e:
                raise ValueError(f"Error loading settings: {e}") from e

        # Handle case where we have valid partial and valid full settings
        self._resolve_full_partial_conflict()
        return self._settings or self._partial_settings

    def _resolve_full_partial_conflict(self):
        """
        Resolve the conflict between full and partial settings, if any.

        Five cases:
        1. No partial settings, no full settings: Nothing to do.
        2. Partial settings, no full settings: Load partial settings.
        3. No partial settings, full settings: Nothing to do.
        4. Partial settings, full settings, and they match: delete partial settings.
        5. Partial settings, full settings, and they don't match: raise ValueError.
        """
        # Handle cases 1 through 3 -- no conflicts in these cases
        if self._partial_settings is None or self._settings is None:
            # Nothing to do, return
            return

        # Both are not None, so try construction full from partial, since partial
        # settings can be full.
        try:
            full_from_partial = PhotometrySettings.model_validate(
                self._partial_settings.model_dump()
            )
        except ValidationError:
            full_from_partial = None

        if full_from_partial != self._settings:
            raise ValueError(
                "Partial settings and full settings do not match. "
                "Please resolve the discrepancy by deleting one of the "
                "settings files."
                f"Folder with settings: {self._working_dir}"
                f"Partial settings: {self._partial_settings_file} "
                f"Full settings: {self._settings_file}"
            )

        # If we reach here, then the partial settings and full settings match, so we
        # can delete the partial settings.
        self._partial_settings_file.unlink()

        # and set the partial settings to None
        self._partial_settings = None



from enum import Enum


class AAVSOFilters(str, Enum):
    """
    The definitive list of AAVSO filters is at https://www.aavso.org/filters
    """

    U = "U"
    B = "B"
    V = "V"
    RJ = "RJ"
    Rc = "R"
    Ic = "I"
    IJ = "IJ"
    J = "J"
    H = "H"
    K = "K"
    TG = "TG"
    TB = "TB"
    TR = "TR"
    CV = "CV"
    CR = "CR"
    SZ = "SZ"
    SU = "SU"
    SG = "SG"
    SR = "SR"
    SI = "SI"
    STU = "STU"
    STV = "STV"
    STB = "STB"
    STY = "STY"
    STHBW = "STHBW"
    STHBN = "STHBN"
    MA = "MA"
    MB = "MB"
    MI = "MI"
    ZS = "ZS"
    Y = "Y"
    HA = "HA"
    HAC = "HAC"
    CBB = "CBB"
    O = "O"  # noqa: E741



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *  # noqa

# ----------------------------------------------------------------------------

from .core import *

# We load this for its side effect of adding YAML representations for the models
from .table_representations import *



import logging
import tempfile
import warnings
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest
from astropy import units as u
from astropy.io import ascii
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.metadata.exceptions import MergeConflictWarning

from stellarphot.core import SourceListData
from stellarphot.photometry import (
    AperturePhotometry,
    calculate_noise,
    find_too_close,
    multi_image_photometry,
    single_image_photometry,
    source_detection,
)
from stellarphot.photometry.tests.fake_image import FakeCCDImage, shift_FakeCCDImage
from stellarphot.settings import (
    Camera,
    LoggingSettings,
    Observatory,
    PassbandMap,
    PhotometryApertures,
    PhotometryOptionalSettings,
    PhotometrySettings,
    SourceLocationSettings,
)

# Constants for the tests

GAINS = [1.0, 1.5, 2.0]
# Make sure the tests are deterministic by using a seed
SEED = 5432985


@pytest.fixture
def source_locations():
    return SourceLocationSettings(
        source_list_file="",
        shift_tolerance=6,
        use_coordinates="pixel",
    )


@pytest.fixture
def photometry_optional_settings():
    pos = PhotometryOptionalSettings()
    # This used to be the default; it has switched to exact
    pos.method = "center"
    return pos


@pytest.fixture
def fake_camera():
    # A camera with not unreasonable settings
    return Camera(
        data_unit=u.adu,
        gain=1.0 * u.electron / u.adu,
        name="test camera",
        read_noise=0 * u.electron,
        dark_current=0.1 * u.electron / u.second,
        pixel_scale=1 * u.arcsec / u.pixel,
        max_data_value=40000 * u.adu,
    )


@pytest.fixture
def zero_camera():
    # Camera with no read noise or dark current
    return Camera(
        data_unit=u.adu,
        gain=1.0 * u.electron / u.adu,
        name="test camera",
        read_noise=0 * u.electron,
        dark_current=0.0 * u.electron / u.second,
        pixel_scale=1 * u.arcsec / u.pixel,
        max_data_value=40000 * u.adu,
    )


@pytest.fixture
def fake_obs():
    # Fake observatory location
    return Observatory(
        name="test observatory",
        latitude=0 * u.deg,
        longitude=0 * u.deg,
        elevation=0 * u.m,
    )


# The fake image used for testing
FAKE_CCD_IMAGE = FakeCCDImage(seed=SEED)


# Build default PhotometryOptions for the tests based on the fake image
@pytest.fixture
def photometry_apertures():
    return PhotometryApertures(
        radius=FAKE_CCD_IMAGE.sources["aperture"][0],
        gap=FAKE_CCD_IMAGE.sources["aperture"][0],
        annulus_width=FAKE_CCD_IMAGE.sources["aperture"][0],
        fwhm=FAKE_CCD_IMAGE.sources["x_stddev"].mean(),
    )


# Passband map for the tests
@pytest.fixture
def passband_map():
    return PassbandMap(
        name="Example Passband Map",
        your_filter_names_to_aavso={
            "B": "B",
            "rp": "SR",
        },
    )


@pytest.fixture
def photometry_settings_for_test(
    fake_camera,
    fake_obs,
    photometry_apertures,
    source_locations,
    photometry_optional_settings,
    passband_map,
):
    return PhotometrySettings(
        camera=fake_camera,
        observatory=fake_obs,
        photometry_apertures=photometry_apertures,
        source_location_settings=source_locations,
        photometry_optional_settings=photometry_optional_settings,
        passband_map=passband_map,
        logging_settings=LoggingSettings(),
    )


class TestAperturePhotometry:
    @staticmethod
    def create_source_list():
        # This has X, Y
        sources = FAKE_CCD_IMAGE.sources.copy()

        # Rename to match the expected names
        sources.rename_column("x_mean", "xcenter")
        sources.rename_column("y_mean", "ycenter")

        # Calculate RA/Dec from image WCS
        coords = FAKE_CCD_IMAGE.wcs.pixel_to_world(
            sources["xcenter"], sources["ycenter"]
        )
        sources["ra"] = coords.ra
        sources["dec"] = coords.dec
        sources["star_id"] = list(range(len(sources)))
        sources["xcenter"] = sources["xcenter"] * u.pixel
        sources["ycenter"] = sources["ycenter"] * u.pixel

        return SourceListData(input_data=sources, colname_map=None)

    @staticmethod
    def list_of_fakes(num_files):
        # Generate fake CCDData objects for use in photometry_on_directory tests
        fake_images = [deepcopy(FAKE_CCD_IMAGE)]

        # Create additional images, each in a different position.
        for i in range(num_files - 1):
            angle = 2 * np.pi / (num_files - 1) * i
            rad = 50
            dx, dy = rad * np.cos(angle), rad * np.sin(angle)
            fake_images.append(shift_FakeCCDImage(fake_images[0], dx, dy))

        filters = ["U", "B", "V", "R", "I"]
        for i in range(num_files):
            if i < 5:
                fake_images[i].header["FILTER"] = filters[i]
            else:
                fake_images[i].header["FILTER"] = "V"

        return fake_images

    def test_create_aperture_photometry(self, tmp_path, photometry_settings_for_test):
        source_list = self.create_source_list()
        source_list_file = tmp_path / "source_list.ecsv"
        source_list.write(source_list_file, overwrite=True)

        # We are fine to modify photometry_settings here because pytest will
        # make a new one for each test.
        photometry_settings_for_test.source_location_settings.source_list_file = str(
            source_list_file
        )

        # Create an AperturePhotometry object
        ap_phot = AperturePhotometry(settings=photometry_settings_for_test)

        # Check that the object was created correctly
        assert ap_phot.settings.camera is photometry_settings_for_test.camera
        assert ap_phot.settings.observatory is photometry_settings_for_test.observatory

    # The True case below is a regression test for #157
    @pytest.mark.parametrize("int_data", [True, False])
    def test_aperture_photometry_no_outlier_rejection(
        self, int_data, tmp_path, photometry_settings_for_test
    ):
        fake_CCDimage = deepcopy(FAKE_CCD_IMAGE)

        found_sources = source_detection(
            fake_CCDimage, fwhm=fake_CCDimage.sources["x_stddev"].mean(), threshold=10
        )

        # The scale_factor is used to rescale data to integers if needed. It
        # needs to be set later on when the net counts are "unscaled" in the
        # asserts that constitute the actual test.
        scale_factor = 1.0
        if int_data:
            scale_factor = (
                0.75
                * photometry_settings_for_test.camera.max_data_value.value
                / fake_CCDimage.data.max()
            )
            # For the moment, ensure the integer data is NOT larger than max_adu
            # because until #161 is fixed then having NaN in the data will not succeed.
            data = scale_factor * fake_CCDimage.data
            fake_CCDimage.data = data.astype(int)

        # Make a copy of photometry options
        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )

        # Modify options to match test before we used phot_options
        phot_options.reject_background_outliers = False
        phot_options.reject_too_close = False
        phot_options.include_dig_noise = True

        source_list_file = tmp_path / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        source_locations = (
            photometry_settings_for_test.source_location_settings.model_copy()
        )
        source_locations.source_list_file = str(source_list_file)

        image_file = tmp_path / "fake_image.fits"
        fake_CCDimage.write(image_file, overwrite=True)

        # We can simply change the fields we want to change now that we use
        # a fixture for photometry_settings.
        photometry_settings_for_test.source_location_settings = source_locations
        photometry_settings_for_test.photometry_optional_settings = phot_options

        ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
        phot, missing_sources = ap_phot(image_file)

        phot.sort("aperture_sum")
        sources = fake_CCDimage.sources
        # Astropy tables sort in-place so we need to sort the sources table
        # after the fact.
        sources.sort("amplitude")
        aperture = photometry_settings_for_test.photometry_apertures.radius

        for inp, out in zip(sources, phot, strict=True):
            stdev = inp["x_stddev"]
            expected_flux = (
                inp["amplitude"]
                * 2
                * np.pi
                * stdev**2
                * (1 - np.exp(-(aperture**2) / (2 * stdev**2)))
            )
            # This expected flux is correct IF there were no noise. With noise, the
            # standard deviation in the sum of the noise within in the aperture is
            # n_pix_in_aperture times the single-pixel standard deviation.

            # We could require that the result be within some reasonable
            # number of those expected variations or we could count up the
            # actual number of background counts at each of the source
            # positions.

            # Here we just check whether any difference is consistent with
            # less than the expected one sigma deviation.

            # We need to remove any scaling that has been done of the data values.
            assert (
                np.abs(expected_flux - out["aperture_net_cnts"].value / scale_factor)
                < np.pi * aperture**2 * fake_CCDimage.noise_dev
            )

    @pytest.mark.parametrize("reject", [True, False])
    def test_aperture_photometry_with_outlier_rejection(
        self, reject, tmp_path, photometry_settings_for_test
    ):
        """
        Insert some really large pixel values in the annulus and check that
        the photometry is correct when outliers are rejected and is
        incorrect when outliers are not rejected.
        """
        fake_CCDimage = deepcopy(FAKE_CCD_IMAGE)
        sources = fake_CCDimage.sources

        aperture_settings = photometry_settings_for_test.photometry_apertures
        aperture = aperture_settings.radius
        inner_annulus = aperture_settings.inner_annulus
        outer_annulus = aperture_settings.outer_annulus

        image = fake_CCDimage.data

        found_sources = source_detection(
            fake_CCDimage, fwhm=sources["x_stddev"].mean(), threshold=10
        )
        source_list_file = tmp_path / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        # Add some large pixel values to the annulus for each source.
        # adding these moves the average pixel value by quite a bit,
        # so we'll only get the correct net flux if these are removed.
        for source in fake_CCDimage.sources:
            center_px = (int(source["x_mean"]), int(source["y_mean"]))
            begin = center_px[0] + inner_annulus + 1
            end = begin + (outer_annulus - inner_annulus - 1)
            # Yes, x and y are deliberately reversed below.
            image[center_px[1], begin:end] = 100 * fake_CCDimage.mean_noise

        # Make a copy of photometry options
        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )

        # Modify options to match test before we used phot_options
        phot_options.reject_background_outliers = reject
        phot_options.reject_too_close = False
        phot_options.include_dig_noise = True

        # It is fine to modify photometry_settings here because pytest will
        # make a new one for each test.
        photometry_settings_for_test.source_location_settings.source_list_file = str(
            source_list_file
        )
        photometry_settings_for_test.photometry_optional_settings = phot_options

        image_file = tmp_path / "fake_image.fits"
        fake_CCDimage.write(image_file, overwrite=True)

        ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
        phot, missing_sources = ap_phot(image_file)

        phot.sort("aperture_sum")
        sources.sort("amplitude")

        for inp, out in zip(sources, phot, strict=True):
            stdev = inp["x_stddev"]
            expected_flux = (
                inp["amplitude"]
                * 2
                * np.pi
                * stdev**2
                * (1 - np.exp(-(aperture**2) / (2 * stdev**2)))
            )
            # This expected flux is correct IF there were no noise. With noise, the
            # standard deviation in the sum of the noise within in the aperture is
            # n_pix_in_aperture times the single-pixel standard deviation.
            #

            expected_deviation = np.pi * aperture**2 * fake_CCDimage.noise_dev
            # We could require that the result be within some reasonable
            # number of those expected variations or we could count up the
            # actual number of background counts at each of the source
            # positions.

            # Here we just check whether any difference is consistent with
            # less than the expected one sigma deviation.
            if reject:
                assert (
                    np.abs(expected_flux - out["aperture_net_cnts"].value)
                    < expected_deviation
                )
            else:
                with pytest.raises(AssertionError):
                    assert (
                        np.abs(expected_flux - out["aperture_net_cnts"].value)
                        < expected_deviation
                    )

    def test_photometry_method_argument(self, tmp_path, photometry_settings_for_test):
        """
        Make sure that setting the method option has an effect on the photometry.
        """
        fake_CCDimage = deepcopy(FAKE_CCD_IMAGE)

        found_sources = source_detection(
            fake_CCDimage, fwhm=fake_CCDimage.sources["x_stddev"].mean(), threshold=10
        )

        source_list_file = tmp_path / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        image_file = tmp_path / "fake_image.fits"
        fake_CCDimage.write(image_file, overwrite=True)

        # Make a copy of photometry options
        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )

        # Do the photometry with method = "center" first
        phot_options.method = "center"

        photometry_settings_for_test.source_location_settings.source_list_file = str(
            source_list_file
        )
        photometry_settings_for_test.photometry_optional_settings = phot_options

        ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
        phot_center, _ = ap_phot(image_file)

        # Now redo photometry with method="exact". Whether the flux is larger or smaller
        # in this case depends on the exact position of the sources in the image.
        # Here we just check that the flux is different.
        phot_options.method = "exact"
        photometry_settings_for_test.photometry_optional_settings = phot_options
        ap_phot_exact = AperturePhotometry(settings=photometry_settings_for_test)
        phot_exact, _ = ap_phot_exact(image_file)
        assert np.all(phot_exact["aperture_sum"] != phot_center["aperture_sum"])

    @pytest.mark.parametrize("coords", ["sky", "pixel"])
    def test_photometry_on_directory(self, coords, photometry_settings_for_test):
        # Create list of fake CCDData objects
        num_files = 5
        fake_images = list_of_fakes(num_files)

        # Write fake images to temporary directory and test
        # multi_image_photometry on them.
        # NOTE: ignore_cleanup_errors=True is needed to avoid an error
        #       when the temporary directory is deleted on Windows.
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            # Come up with Filenames
            temp_file_names = [
                Path(temp_dir) / f"tempfile_{i:02d}.fit"
                for i in range(1, num_files + 1)
            ]
            # Write the CCDData objects to files
            for i, image in enumerate(fake_images):
                image.write(temp_file_names[i])

            object_name = fake_images[0].header["OBJECT"]
            sources = fake_images[0].sources
            aperture_settings = photometry_settings_for_test.photometry_apertures
            aperture = aperture_settings.radius

            # Generate the sourcelist
            found_sources = source_detection(
                fake_images[0],
                fwhm=fake_images[0].sources["x_stddev"].mean(),
                threshold=10,
            )

            source_list_file = Path(temp_dir) / "source_list.ecsv"
            found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

            # Make a copy of photometry options
            phot_options = (
                photometry_settings_for_test.photometry_optional_settings.model_copy()
            )

            # Modify options to match test before we used phot_options
            phot_options.include_dig_noise = True
            phot_options.reject_too_close = True
            phot_options.reject_background_outliers = True
            phot_options.fwhm_by_fit = True

            photometry_settings_for_test.photometry_optional_settings = phot_options
            photometry_settings_for_test.source_location_settings.use_coordinates = (
                coords
            )
            photometry_settings_for_test.source_location_settings.source_list_file = (
                str(source_list_file)
            )
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message="Cannot merge meta key",
                    category=MergeConflictWarning,
                )
                ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
                phot_data = ap_phot(
                    temp_dir,
                    object_of_interest=object_name,
                )

        # For following assertion to be true, rad must be small enough that
        # no source lies within outer_annulus of the edge of an image.
        assert len(phot_data) == num_files * len(found_sources)

        # Sort all data by amount of signal
        sources.sort("amplitude")
        found_sources.sort("flux")

        # Get noise level from the first image
        noise_dev = fake_images[0].noise_dev

        for fnd, inp in zip(found_sources, sources, strict=True):
            star_id_chk = fnd["star_id"]
            # Select the rows in phot_data that correspond to the current star
            # and compute the average of the aperture sums.
            selected_rows = phot_data[phot_data["star_id"] == star_id_chk]
            obs_avg_net_cnts = np.average(selected_rows["aperture_net_cnts"].value)

            stdev = inp["x_stddev"]
            expected_flux = (
                inp["amplitude"]
                * 2
                * np.pi
                * stdev**2
                * (1 - np.exp(-(aperture**2) / (2 * stdev**2)))
            )
            # This expected flux is correct IF there were no noise. With noise, the
            # standard deviation in the sum of the noise within in the aperture is
            # n_pix_in_aperture times the single-pixel standard deviation.
            #

            expected_deviation = np.pi * aperture**2 * noise_dev

            # We have two cases to consider: use_coordinates="sky" and
            # use_coordinates="pixel".
            if coords == "sky":
                # In this case, the expected result is the test below.

                # We could require that the result be within some reasonable
                # number of those expected variations or we could count up the
                # actual number of background counts at each of the source
                # positions.

                # Here we just check whether any difference is consistent with
                # less than the expected one sigma deviation.
                assert np.abs(expected_flux - obs_avg_net_cnts) < expected_deviation
            else:
                # In this case we are trying to do photometry in pixel coordinates,
                # using the pixel location of the sources as found in the first image --
                # see the line where found_sources is defined.
                #
                # However, the images are shifted with respect to each other by
                # list_of_fakes, so there are no longer stars at those positions in the
                # other images.
                #
                # Because of that, the expected result is that either obs_avg_net_cnts
                # is nan or the difference is bigger than the expected_deviation.
                assert (
                    np.isnan(obs_avg_net_cnts)
                    or np.abs(expected_flux - obs_avg_net_cnts) > expected_deviation
                )

    def test_photometry_on_directory_with_no_ra_dec(self, photometry_settings_for_test):
        # Create list of fake CCDData objects
        num_files = 5
        fake_images = list_of_fakes(num_files)

        # Write fake images to temporary directory and test
        # multi_image_photometry on them.
        # NOTE: ignore_cleanup_errors=True is needed to avoid an error
        #       when the temporary directory is deleted on Windows.
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            # Come up with Filenames
            temp_file_names = [
                Path(temp_dir) / f"tempfile_{i:02d}.fits"
                for i in range(1, num_files + 1)
            ]
            # Write the CCDData objects to files
            for i, image in enumerate(fake_images):
                image.write(temp_file_names[i])

            object_name = fake_images[0].header["OBJECT"]

            # Generate the sourcelist
            found_sources = source_detection(
                fake_images[0],
                fwhm=fake_images[0].sources["x_stddev"].mean(),
                threshold=10,
            )

            # Damage the sourcelist by removing the ra and dec columns
            found_sources.drop_ra_dec()

            source_list_file = Path(temp_dir) / "source_list.ecsv"
            found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

            phot_options = (
                photometry_settings_for_test.photometry_optional_settings.model_copy()
            )

            # Modify options to match test before we used phot_options
            phot_options.include_dig_noise = True
            phot_options.reject_too_close = True
            phot_options.reject_background_outliers = True
            phot_options.fwhm_by_fit = True

            photometry_settings_for_test.photometry_optional_settings = phot_options
            photometry_settings_for_test.source_location_settings.source_list_file = (
                str(source_list_file)
            )
            # The setting below was implicit in the old default
            photometry_settings_for_test.source_location_settings.use_coordinates = (
                "sky"
            )

            ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
            with pytest.raises(ValueError):
                ap_phot(
                    temp_dir,
                    object_of_interest=object_name,
                )

    def test_photometry_on_directory_with_bad_fits(self, photometry_settings_for_test):
        # Create list of fake CCDData objects
        num_files = 5
        clean_fake_images = list_of_fakes(num_files)
        fake_images = list_of_fakes(num_files)

        # Write fake images (without WCS) to temporary directory and test
        # multi_image_photometry on them.
        # NOTE: ignore_cleanup_errors=True is needed to avoid an error
        #       when the temporary directory is deleted on Windows.
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            # Come up with Filenames
            temp_file_names = [
                Path(temp_dir) / f"tempfile_{i:02d}.fits"
                for i in range(1, num_files + 1)
            ]
            # Write the CCDData objects to files
            for i, image in enumerate(fake_images):
                image.drop_wcs()
                image.write(temp_file_names[i])

            object_name = fake_images[0].header["OBJECT"]

            # Generate the sourcelist with RA/Dec information from a clean image
            found_sources = source_detection(
                clean_fake_images[0],
                fwhm=clean_fake_images[0].sources["x_stddev"].mean(),
                threshold=10,
            )

            source_list_file = Path(temp_dir) / "source_list.ecsv"
            found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

            phot_options = (
                photometry_settings_for_test.photometry_optional_settings.model_copy()
            )

            # Modify options to match test before we used phot_options
            phot_options.include_dig_noise = True
            phot_options.reject_too_close = True
            phot_options.reject_background_outliers = True
            phot_options.fwhm_by_fit = True

            photometry_settings_for_test.photometry_optional_settings = phot_options
            photometry_settings_for_test.source_location_settings.source_list_file = (
                str(source_list_file)
            )
            # The settings below was implicit in the old default
            photometry_settings_for_test.source_location_settings.use_coordinates = (
                "sky"
            )

            ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
            # Since none of the images will be valid, it should raise a RuntimeError
            with pytest.raises(RuntimeError):
                ap_phot(
                    temp_dir,
                    object_of_interest=object_name,
                )

    def test_invalid_path(self, photometry_settings_for_test):
        ap = AperturePhotometry(settings=photometry_settings_for_test)
        with pytest.raises(ValueError, match="is not a valid file or directory"):
            ap("invalid_path")

    # Checking logging for AperturePhotometry for single image photometry.
    @pytest.mark.parametrize("logfile", ["test.log", None])
    @pytest.mark.parametrize("console_log", [True, False])
    def test_logging_single_image(
        self, capsys, logfile, console_log, tmp_path, photometry_settings_for_test
    ):
        # Disable any root logger handlers that are active before using
        # logging since that is expectation of single_image_photometry.
        if logging.root.hasHandlers():
            logging.root.handlers.clear()

        # Create fake image
        fake_CCDimage = deepcopy(FAKE_CCD_IMAGE)
        image_file = tmp_path / "fake_image.fits"
        fake_CCDimage.write(image_file, overwrite=True)
        # Create source list from fake image
        found_sources = source_detection(
            fake_CCDimage, fwhm=fake_CCDimage.sources["x_stddev"].mean(), threshold=10
        )
        source_list_file = tmp_path / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        # Make a copy of photometry options and modify them to match the
        # test_aperture_photometry_no_outlier_rejection settings
        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )
        phot_options.reject_background_outliers = False
        phot_options.reject_too_close = False
        phot_options.include_dig_noise = True

        # Define the source locations settings
        source_locations = (
            photometry_settings_for_test.source_location_settings.model_copy()
        )
        source_locations.source_list_file = str(source_list_file)

        # Define the logging settings
        logging_settings = photometry_settings_for_test.logging_settings.model_copy()
        if logfile:
            # Define the log file and console log settings
            # and make sure to set full path of log file.
            logging_settings.logfile = str(tmp_path / logfile)
            logging_settings.console_log = console_log
            full_logfile = logging_settings.logfile

        photometry_settings_for_test.source_location_settings = source_locations
        photometry_settings_for_test.photometry_optional_settings = phot_options
        photometry_settings_for_test.logging_settings = logging_settings

        # Call the AperturePhotometry class with a single image
        ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
        phot, missing_sources = ap_phot(image_file)

        #
        # Test logging was consistent with settings
        #
        # Check and see if the output log file was created and contains the
        # expected messages.
        if logfile:
            assert Path(full_logfile).exists()
            with open(full_logfile) as f:
                log_content = f.read()
                # Confirm last log message written by single_image_photometry
                # present.
                assert "Calculating noise for all sources" in log_content

        # If console logging is enabled then the stderr should contain the
        # expected messages.
        if console_log:
            captured_stdout = capsys.readouterr()
            # Confirm last log message written by single_image_photometry
            # present.
            assert "Calculating noise for all sources" in captured_stdout.err

    # Checking logging for AperturePhotometry for multiple image photometry.
    @pytest.mark.parametrize("logfile", ["test.log", None])
    @pytest.mark.parametrize("console_log", [True, False])
    def test_logging_multiple_image(
        self, capsys, logfile, console_log, photometry_settings_for_test
    ):
        # Create list of fake CCDData objects
        num_files = 5
        fake_images = list_of_fakes(num_files)

        # Write fake images to temporary directory and test
        # multi_image_photometry on them.0
        # NOTE: ignore_cleanup_errors=True is needed to avoid an error
        #       when the temporary directory is deleted on Windows.
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            # Come up with Filenames
            temp_file_names = [
                Path(temp_dir) / f"tempfile_{i:02d}.fit"
                for i in range(1, num_files + 1)
            ]
            # Write the CCDData objects to files
            for i, image in enumerate(fake_images):
                image.write(temp_file_names[i])
            object_name = fake_images[0].header["OBJECT"]

            # Generate the sourcelist
            found_sources = source_detection(
                fake_images[0],
                fwhm=fake_images[0].sources["x_stddev"].mean(),
                threshold=10,
            )
            source_list_file = Path(temp_dir) / "source_list.ecsv"
            found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

            # Make a copy of photometry options based on those used in
            # successful test_photometry_on_directory
            phot_options = (
                photometry_settings_for_test.photometry_optional_settings.model_copy()
            )
            phot_options.include_dig_noise = True
            phot_options.reject_too_close = True
            phot_options.reject_background_outliers = True
            phot_options.fwhm_by_fit = True

            photometry_settings_for_test.photometry_optional_settings = phot_options
            photometry_settings_for_test.source_location_settings.use_coordinates = (
                "sky"
            )
            photometry_settings_for_test.source_location_settings.source_list_file = (
                str(source_list_file)
            )

            # Define the logging settings
            logging_settings = (
                photometry_settings_for_test.logging_settings.model_copy()
            )
            if logfile:
                logging_settings.logfile = logfile
                logging_settings.console_log = console_log
                # log file should be written to image directory (tmp_dir)
                # automatically, this ensures we know the path to the log file.
                full_logfile = str(Path(temp_dir) / logfile)
                photometry_settings_for_test.logging_settings = logging_settings

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message="Cannot merge meta key",
                    category=MergeConflictWarning,
                )

                ap_phot = AperturePhotometry(settings=photometry_settings_for_test)
                _ = ap_phot(temp_dir, object_of_interest=object_name)

                #
                # Test logging was consistent with settings
                #
                # Check and see if the output log file was created and contains the
                # expected messages.
                if logfile:
                    assert Path(full_logfile).exists()
                    with open(full_logfile) as f:
                        log_content = f.read()
                        # Check for log messages output by multi_image_photometry
                        assert "Starting photometry of files in" in log_content
                        assert "DONE processing all matching images" in log_content
                        # Check for last log message from single_image_photometry
                        assert "Calculating noise for all sources" in log_content

                # If console logging is enabled then the stderr should contain the
                # expected messages.
                if console_log:
                    captured_stdout = capsys.readouterr()
                    # Check for log messages output by multi_image_photometry
                    assert "Starting photometry of files in" in captured_stdout.err
                    assert "DONE processing all matching images" in captured_stdout.err
                    # Check for last log message from single_image_photometry
                    assert "Calculating noise for all sources" in captured_stdout.err


def test_calc_noise_defaults():
    # If we put in nothing we should get an error about is missing camera
    # instance.
    with pytest.raises(ValueError):
        assert calculate_noise() == 0


@pytest.mark.parametrize("aperture_area", [5, 20])
@pytest.mark.parametrize("gain", GAINS)
def test_calc_noise_source_only(gain, aperture_area, zero_camera):
    # If the only source of noise is Poisson error in the source
    # then the noise should be the square root of the counts.
    counts = 100
    expected = np.sqrt(gain * counts)

    # Create camera instance
    camera = zero_camera.model_copy()
    camera.gain = gain * camera.gain.unit

    np.testing.assert_allclose(
        calculate_noise(camera, counts=counts, aperture_area=aperture_area), expected
    )


@pytest.mark.parametrize("aperture_area", [5, 20])
@pytest.mark.parametrize("gain", GAINS)
def test_calc_noise_dark_only(gain, aperture_area, zero_camera):
    # Gain should not affect this one. Dark current needs a couple other things,
    # but this is basically Poisson error.
    dark_current = 10
    exposure = 20

    # Create camera instance
    camera = zero_camera.model_copy()
    # Set gain and dark current to values for test
    camera.dark_current = dark_current * camera.dark_current.unit
    camera.gain = gain * camera.gain.unit

    expected = np.sqrt(dark_current * aperture_area * exposure)

    np.testing.assert_allclose(
        calculate_noise(camera, aperture_area=aperture_area, exposure=exposure),
        expected,
    )


@pytest.mark.parametrize("aperture_area", [5, 20])
@pytest.mark.parametrize("gain", GAINS)
def test_calc_read_noise_only(gain, aperture_area, zero_camera):
    # The read noise per pixel IS the noise. The only multiplier is
    # the number of pixels.
    read_noise = 10
    expected = np.sqrt(aperture_area * read_noise**2)

    # Create camera instance
    camera = zero_camera.model_copy()
    camera.read_noise = read_noise * camera.read_noise.unit
    camera.gain = gain * camera.gain.unit

    np.testing.assert_allclose(
        calculate_noise(camera, aperture_area=aperture_area), expected
    )


@pytest.mark.parametrize("aperture_area", [5, 20])
@pytest.mark.parametrize("gain", GAINS)
def test_calc_sky_only(gain, aperture_area, zero_camera):
    # The sky noise per pixel is the poisson and per pixel.
    sky = 10
    expected = np.sqrt(gain * aperture_area * sky)

    # Create camera instance
    camera = zero_camera.model_copy()
    camera.gain = gain * camera.gain.unit

    np.testing.assert_allclose(
        calculate_noise(camera, aperture_area=aperture_area, sky_per_pix=sky), expected
    )


def test_annulus_area_term(zero_camera):
    # Test that noise is correct with an annulus
    aperture_area = 20

    # Annulus is typically quite a bit larger than aperture.
    annulus_area = 10 * aperture_area
    gain = 1.5
    sky = 10
    expected = np.sqrt(gain * aperture_area * (1 + aperture_area / annulus_area) * sky)

    # Create camera instance
    camera = zero_camera.model_copy()
    camera.gain = gain * camera.gain.unit

    np.testing.assert_allclose(
        calculate_noise(
            camera,
            aperture_area=aperture_area,
            annulus_area=annulus_area,
            sky_per_pix=sky,
        ),
        expected,
    )


@pytest.mark.parametrize("digit,expected", ((False, 89.078616), (True, 89.10182)))
def test_calc_noise_messy_case(digit, expected, zero_camera):
    # Do a single test where all the parameters are set and compare with
    # what a calculator gave.
    counts = 1000

    aperture_area = 20
    annulus_area = 10 * aperture_area

    gain = 1.5
    sky = 15
    dark_current = 7
    exposure = 18
    read_noise = 12

    # Create camera instance
    camera = zero_camera.model_copy()
    camera.gain = gain * camera.gain.unit
    camera.dark_current = dark_current * camera.dark_current.unit
    camera.read_noise = read_noise * camera.read_noise.unit

    np.testing.assert_allclose(
        calculate_noise(
            camera,
            counts=counts,
            sky_per_pix=sky,
            exposure=exposure,
            aperture_area=aperture_area,
            annulus_area=annulus_area,
            include_digitization=digit,
        ),
        expected,
    )


def test_find_too_close():
    # Load test sourcelist into memory
    test_sl_data = ascii.read(
        get_pkg_data_filename("data/test_corner.ecsv"), format="ecsv", fast_reader=False
    )

    # Create no sky position sourcelist
    test_sl_data_nosky = test_sl_data.copy()
    test_sl_data_nosky.remove_column("ra")
    test_sl_data_nosky.remove_column("dec")

    # Create no image position sourcelist
    test_sl_data_noimgpos = test_sl_data.copy()
    test_sl_data_noimgpos.remove_column("xcenter")
    test_sl_data_noimgpos.remove_column("ycenter")

    # Create SourceListData objects
    sl_test = SourceListData(input_data=test_sl_data, colname_map=None)
    sl_test_nosky = SourceListData(input_data=test_sl_data_nosky, colname_map=None)
    sl_test_noimgpos = SourceListData(
        input_data=test_sl_data_noimgpos, colname_map=None
    )

    assert sl_test.has_ra_dec is True
    assert sl_test.has_x_y is True
    assert sl_test_nosky.has_ra_dec is False
    assert sl_test_nosky.has_x_y is True
    assert sl_test_noimgpos.has_ra_dec is True
    assert sl_test_noimgpos.has_x_y is False

    # Test full positions available
    ap_in_asec = 5
    feder_scale = 0.563
    aperture_rad = ap_in_asec / feder_scale

    rejects = find_too_close(sl_test, aperture_rad, pixel_scale=feder_scale)
    assert np.sum(rejects) == 5

    # Test only sky positions available
    rejects = find_too_close(sl_test_noimgpos, aperture_rad, pixel_scale=feder_scale)
    assert np.sum(rejects) == 5

    # Test only image positions available
    rejects = find_too_close(sl_test_nosky, aperture_rad, pixel_scale=feder_scale)
    assert np.sum(rejects) == 5


# The True case below is a regression test for #157
@pytest.mark.parametrize("int_data", [True, False])
def test_aperture_photometry_no_outlier_rejection(
    int_data, tmp_path, photometry_settings_for_test
):
    fake_CCDimage = deepcopy(FAKE_CCD_IMAGE)

    found_sources = source_detection(
        fake_CCDimage, fwhm=fake_CCDimage.sources["x_stddev"].mean(), threshold=10
    )

    # The scale_factor is used to rescale data to integers if needed. It
    # needs to be set later on when the net counts are "unscaled" in the
    # asserts that constitute the actual test.
    scale_factor = 1.0
    if int_data:
        scale_factor = (
            0.75
            * photometry_settings_for_test.camera.max_data_value.value
            / fake_CCDimage.data.max()
        )
        # For the moment, ensure the integer data is NOT larger than max_adu
        # because until #161 is fixed then having NaN in the data will not succeed.
        data = scale_factor * fake_CCDimage.data
        fake_CCDimage.data = data.astype(int)

    # Make a copy of photometry options
    phot_options = (
        photometry_settings_for_test.photometry_optional_settings.model_copy()
    )

    # Modify options to match test before we used phot_options
    phot_options.reject_background_outliers = False
    phot_options.reject_too_close = False
    phot_options.include_dig_noise = True

    source_list_file = tmp_path / "source_list.ecsv"
    found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

    source_locations = (
        photometry_settings_for_test.source_location_settings.model_copy()
    )
    source_locations.source_list_file = str(source_list_file)

    # Customize a couple options for this test
    photometry_settings_for_test.source_location_settings = source_locations
    photometry_settings_for_test.photometry_optional_settings = phot_options

    phot, missing_sources = single_image_photometry(
        fake_CCDimage,
        photometry_settings_for_test,
    )

    phot.sort("aperture_sum")
    sources = fake_CCDimage.sources
    # Astropy tables sort in-place so we need to sort the sources table
    # after the fact.
    sources.sort("amplitude")
    aperture = photometry_settings_for_test.photometry_apertures.radius

    for inp, out in zip(sources, phot, strict=True):
        stdev = inp["x_stddev"]
        expected_flux = (
            inp["amplitude"]
            * 2
            * np.pi
            * stdev**2
            * (1 - np.exp(-(aperture**2) / (2 * stdev**2)))
        )
        # This expected flux is correct IF there were no noise. With noise, the
        # standard deviation in the sum of the noise within in the aperture is
        # n_pix_in_aperture times the single-pixel standard deviation.

        # We could require that the result be within some reasonable
        # number of those expected variations or we could count up the
        # actual number of background counts at each of the source
        # positions.

        # Here we just check whether any difference is consistent with
        # less than the expected one sigma deviation.

        # We need to remove any scaling that has been done of the data values.
        assert (
            np.abs(expected_flux - out["aperture_net_cnts"].value / scale_factor)
            < np.pi * aperture**2 * fake_CCDimage.noise_dev
        )


@pytest.mark.parametrize("reject", [True, False])
def test_aperture_photometry_with_outlier_rejection(
    reject, tmp_path, photometry_settings_for_test
):
    """
    Insert some really large pixel values in the annulus and check that
    the photometry is correct when outliers are rejected and is
    incorrect when outliers are not rejected.
    """
    fake_CCDimage = deepcopy(FAKE_CCD_IMAGE)
    sources = fake_CCDimage.sources

    aperture_settings = photometry_settings_for_test.photometry_apertures
    aperture = aperture_settings.radius
    inner_annulus = aperture_settings.inner_annulus
    outer_annulus = aperture_settings.outer_annulus

    image = fake_CCDimage.data

    found_sources = source_detection(
        fake_CCDimage, fwhm=sources["x_stddev"].mean(), threshold=10
    )
    source_list_file = tmp_path / "source_list.ecsv"
    found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

    # Add some large pixel values to the annulus for each source.
    # adding these moves the average pixel value by quite a bit,
    # so we'll only get the correct net flux if these are removed.
    for source in fake_CCDimage.sources:
        center_px = (int(source["x_mean"]), int(source["y_mean"]))
        begin = center_px[0] + inner_annulus + 1
        end = begin + (outer_annulus - inner_annulus - 1)
        # Yes, x and y are deliberately reversed below.
        image[center_px[1], begin:end] = 100 * fake_CCDimage.mean_noise

    # Make a copy of photometry options
    phot_options = (
        photometry_settings_for_test.photometry_optional_settings.model_copy()
    )

    # Modify options to match test before we used phot_options
    phot_options.reject_background_outliers = reject
    phot_options.reject_too_close = False
    phot_options.include_dig_noise = True

    photometry_settings_for_test.source_location_settings.source_list_file = str(
        source_list_file
    )
    photometry_settings_for_test.photometry_optional_settings = phot_options

    phot, missing_sources = single_image_photometry(
        fake_CCDimage,
        photometry_settings_for_test,
    )

    phot.sort("aperture_sum")
    sources.sort("amplitude")

    for inp, out in zip(sources, phot, strict=True):
        stdev = inp["x_stddev"]
        expected_flux = (
            inp["amplitude"]
            * 2
            * np.pi
            * stdev**2
            * (1 - np.exp(-(aperture**2) / (2 * stdev**2)))
        )
        # This expected flux is correct IF there were no noise. With noise, the
        # standard deviation in the sum of the noise within in the aperture is
        # n_pix_in_aperture times the single-pixel standard deviation.
        #

        expected_deviation = np.pi * aperture**2 * fake_CCDimage.noise_dev
        # We could require that the result be within some reasonable
        # number of those expected variations or we could count up the
        # actual number of background counts at each of the source
        # positions.

        # Here we just check whether any difference is consistent with
        # less than the expected one sigma deviation.
        if reject:
            assert (
                np.abs(expected_flux - out["aperture_net_cnts"].value)
                < expected_deviation
            )
        else:
            with pytest.raises(AssertionError):
                assert (
                    np.abs(expected_flux - out["aperture_net_cnts"].value)
                    < expected_deviation
                )


def list_of_fakes(num_files):
    # Generate fake CCDData objects for use in photometry_on_directory tests
    fake_images = [deepcopy(FAKE_CCD_IMAGE)]

    # Create additional images, each in a different position.
    for i in range(num_files - 1):
        angle = 2 * np.pi / (num_files - 1) * i
        rad = 50
        dx, dy = rad * np.cos(angle), rad * np.sin(angle)
        fake_images.append(shift_FakeCCDImage(fake_images[0], dx, dy))

    filters = ["U", "B", "V", "R", "I"]
    for i in range(num_files):
        if i < 5:
            fake_images[i].header["FILTER"] = filters[i]
        else:
            fake_images[i].header["FILTER"] = "V"

    return fake_images


@pytest.mark.parametrize("coords", ["sky", "pixel"])
def test_photometry_on_directory(coords, photometry_settings_for_test):
    # Create list of fake CCDData objects
    num_files = 5
    fake_images = list_of_fakes(num_files)

    # Write fake images to temporary directory and test
    # multi_image_photometry on them.
    # NOTE: ignore_cleanup_errors=True is needed to avoid an error
    #       when the temporary directory is deleted on Windows.
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
        # Come up with Filenames
        temp_file_names = [
            Path(temp_dir) / f"tempfile_{i:02d}.fit" for i in range(1, num_files + 1)
        ]
        # Write the CCDData objects to files
        for i, image in enumerate(fake_images):
            image.write(temp_file_names[i])

        object_name = fake_images[0].header["OBJECT"]
        sources = fake_images[0].sources
        aperture_settings = photometry_settings_for_test.photometry_apertures
        aperture = aperture_settings.radius

        # Generate the sourcelist
        found_sources = source_detection(
            fake_images[0], fwhm=fake_images[0].sources["x_stddev"].mean(), threshold=10
        )

        source_list_file = Path(temp_dir) / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        # Make a copy of photometry options
        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )

        # Modify options to match test before we used phot_options
        phot_options.include_dig_noise = True
        phot_options.reject_too_close = True
        phot_options.reject_background_outliers = True
        phot_options.fwhm_by_fit = True

        photometry_settings_for_test.photometry_optional_settings = phot_options
        photometry_settings_for_test.source_location_settings.use_coordinates = coords
        photometry_settings_for_test.source_location_settings.source_list_file = str(
            source_list_file
        )
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="Cannot merge meta key", category=MergeConflictWarning
            )
            phot_data = multi_image_photometry(
                temp_dir,
                photometry_settings_for_test,
                object_of_interest=object_name,
            )

    # For following assertion to be true, rad must be small enough that
    # no source lies within outer_annulus of the edge of an image.
    assert len(phot_data) == num_files * len(found_sources)

    # Sort all data by amount of signal
    sources.sort("amplitude")
    found_sources.sort("flux")

    # Get noise level from the first image
    noise_dev = fake_images[0].noise_dev

    for fnd, inp in zip(found_sources, sources, strict=True):
        star_id_chk = fnd["star_id"]
        # Select the rows in phot_data that correspond to the current star
        # and compute the average of the aperture sums.
        selected_rows = phot_data[phot_data["star_id"] == star_id_chk]
        obs_avg_net_cnts = np.average(selected_rows["aperture_net_cnts"].value)

        stdev = inp["x_stddev"]
        expected_flux = (
            inp["amplitude"]
            * 2
            * np.pi
            * stdev**2
            * (1 - np.exp(-(aperture**2) / (2 * stdev**2)))
        )
        # This expected flux is correct IF there were no noise. With noise, the
        # standard deviation in the sum of the noise within in the aperture is
        # n_pix_in_aperture times the single-pixel standard deviation.
        #

        expected_deviation = np.pi * aperture**2 * noise_dev

        # We have two cases to consider: use_coordinates="sky" and
        # use_coordinates="pixel".
        if coords == "sky":
            # In this case, the expected result is the test below.

            # We could require that the result be within some reasonable
            # number of those expected variations or we could count up the
            # actual number of background counts at each of the source
            # positions.

            # Here we just check whether any difference is consistent with
            # less than the expected one sigma deviation.
            assert np.abs(expected_flux - obs_avg_net_cnts) < expected_deviation
        else:
            # In this case we are trying to do photometry in pixel coordinates,
            # using the pixel location of the sources as found in the first image --
            # see the line where found_sources is defined.
            #
            # However, the images are shifted with respect to each other by
            # list_of_fakes, so there are no longer stars at those positions in the
            # other images.
            #
            # Because of that, the expected result is that either obs_avg_net_cnts
            # is nan or the difference is bigger than the expected_deviation.
            assert (
                np.isnan(obs_avg_net_cnts)
                or np.abs(expected_flux - obs_avg_net_cnts) > expected_deviation
            )


def test_photometry_on_directory_with_no_ra_dec(photometry_settings_for_test):
    # Create list of fake CCDData objects
    num_files = 5
    fake_images = list_of_fakes(num_files)

    # Write fake images to temporary directory and test
    # multi_image_photometry on them.
    # NOTE: ignore_cleanup_errors=True is needed to avoid an error
    #       when the temporary directory is deleted on Windows.
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
        # Come up with Filenames
        temp_file_names = [
            Path(temp_dir) / f"tempfile_{i:02d}.fits" for i in range(1, num_files + 1)
        ]
        # Write the CCDData objects to files
        for i, image in enumerate(fake_images):
            image.write(temp_file_names[i])

        object_name = fake_images[0].header["OBJECT"]

        # Generate the sourcelist
        found_sources = source_detection(
            fake_images[0], fwhm=fake_images[0].sources["x_stddev"].mean(), threshold=10
        )

        # Damage the sourcelist by removing the ra and dec columns
        found_sources.drop_ra_dec()

        source_list_file = Path(temp_dir) / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )

        # Modify options to match test before we used phot_options
        phot_options.include_dig_noise = True
        phot_options.reject_too_close = True
        phot_options.reject_background_outliers = True
        phot_options.fwhm_by_fit = True

        photometry_settings_for_test.photometry_optional_settings = phot_options
        photometry_settings_for_test.source_location_settings.source_list_file = str(
            source_list_file
        )
        # The setting below was implicit in the old default
        photometry_settings_for_test.source_location_settings.use_coordinates = "sky"

        with pytest.raises(ValueError):
            multi_image_photometry(
                temp_dir,
                photometry_settings_for_test,
                object_of_interest=object_name,
            )


def test_photometry_on_directory_with_bad_fits(photometry_settings_for_test):
    # Create list of fake CCDData objects
    num_files = 5
    clean_fake_images = list_of_fakes(num_files)
    fake_images = list_of_fakes(num_files)

    # Write fake images (without WCS) to temporary directory and test
    # multi_image_photometry on them.
    # NOTE: ignore_cleanup_errors=True is needed to avoid an error
    #       when the temporary directory is deleted on Windows.
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
        # Come up with Filenames
        temp_file_names = [
            Path(temp_dir) / f"tempfile_{i:02d}.fits" for i in range(1, num_files + 1)
        ]
        # Write the CCDData objects to files
        for i, image in enumerate(fake_images):
            image.drop_wcs()
            image.write(temp_file_names[i])

        object_name = fake_images[0].header["OBJECT"]

        # Generate the sourcelist with RA/Dec information from a clean image
        found_sources = source_detection(
            clean_fake_images[0],
            fwhm=clean_fake_images[0].sources["x_stddev"].mean(),
            threshold=10,
        )

        source_list_file = Path(temp_dir) / "source_list.ecsv"
        found_sources.write(source_list_file, format="ascii.ecsv", overwrite=True)

        phot_options = (
            photometry_settings_for_test.photometry_optional_settings.model_copy()
        )

        # Modify options to match test before we used phot_options
        phot_options.include_dig_noise = True
        phot_options.reject_too_close = True
        phot_options.reject_background_outliers = True
        phot_options.fwhm_by_fit = True

        photometry_settings_for_test.photometry_optional_settings = phot_options
        photometry_settings_for_test.source_location_settings.source_list_file = str(
            source_list_file
        )
        # The settings below was implicit in the old default
        photometry_settings_for_test.source_location_settings.use_coordinates = "sky"

        # Since none of the images will be valid, it should raise a RuntimeError
        with pytest.raises(RuntimeError):
            multi_image_photometry(
                temp_dir,
                photometry_settings_for_test,
                object_of_interest=object_name,
            )



import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import QTable, Table

__all__ = ["add_in_quadrature", "calc_aij_relative_flux"]


def add_in_quadrature(array):
    """
    Add an array of numbers in quadrature.
    """
    return np.sqrt((array**2).sum())


def calc_aij_relative_flux(
    star_data,
    comp_stars,
    in_place=True,
    coord_column=None,
    star_id_column="star_id",
    counts_column_name="aperture_net_cnts",
):
    """
    Calculate AstroImageJ-style flux ratios.

    Parameters
    ----------

    star_data : 'stellarphot.PhotometryData'
        Photometry data from one or more images.

    comp_stars : '~astropy.table.Table'
        Table of comparison stars in the field. Must contain a column
        called ``ra`` and a column called ``dec``.
        NOTE that not all
        of the comparison stars will necessarily be used. Stars in
        this table are excluded from the comparison set if, in any
        of the `star_data` for that comparison, the net counts are
        ``NaN`` or if the angular distance between the position in
        the `star_data` and the position in the `comp_stars` table
        is too large.

    in_place : bool,  optional
        If ``True``, add new columns to input table. Otherwise, return
        new table with those columns added.

    coord_column : str,  optional
        If provided, use this column to match comparison stars to coordinates.
        If not provided, the coordinates are generated with SkyCoord.

    counts_column_name : str,  optional
        If provided, use this column to find counts.

    star_id_column : str,  optional
        Name of the column that provides a unique identifier for each
        comparison star.

    Returns
    -------

    `stellarphot.PhotometryData` or None
        The return type depends on the value of ``in_place``. If it is
        ``False``, then the new columns are returned as a separate table,
        otherwise the columns are simply added to the input table.
    """

    # Match comparison star list to instrumental magnitude information
    if star_data["ra"].unit is None:
        unit = "degree"
    else:
        # Pulled this from the source code -- None is ok but need
        # to match the number of coordinates.
        unit = [None, None]

    star_data_coords = SkyCoord(ra=star_data["ra"], dec=star_data["dec"], unit=unit)

    if coord_column is not None:
        comp_coords = comp_stars[coord_column]
    else:
        if comp_stars["ra"].unit is None:
            unit = "degree"
        else:
            # Pulled this from the source code -- None is ok but need
            # to match the number of coordinates.
            unit = [None, None]
        comp_coords = SkyCoord(ra=comp_stars["ra"], dec=comp_stars["dec"], unit=unit)

    # Check for matches of stars in star data to the stars in comp_stars
    # and eliminate as comps any stars for which the separation is bigger
    # than 1.2 arcsec in any of the frames.
    index, d2d, _ = star_data_coords.match_to_catalog_sky(comp_coords)

    # Not sure this is really close enough for a good match...
    good = d2d < 1.2 * u.arcsec

    check_for_bad = Table(
        data=[star_data[star_id_column].data, good], names=["star_id", "good"]
    )
    check_for_bad = check_for_bad.group_by("star_id")
    is_all_good = check_for_bad.groups.aggregate(np.all)

    bad_comps = set(is_all_good["star_id"][~is_all_good["good"]])

    # Check whether any of the comp stars have NaN values and,
    # if they do, exclude them from the comp set.
    check_for_nan = Table(
        data=[star_data[star_id_column].data, star_data[counts_column_name].data],
        names=["star_id", "net_counts"],
    )
    check_for_nan = check_for_nan.group_by("star_id")
    check_for_nan["good"] = ~np.isnan(check_for_nan["net_counts"])
    is_all_good = check_for_nan.groups.aggregate(np.all)

    bad_comps = bad_comps | set(is_all_good["star_id"][~is_all_good["good"]])

    for comp in bad_comps:
        this_comp = star_data[star_id_column] == comp
        good[this_comp] = False

    error_column_name = "noise_electrons"
    # Calculate comp star counts for each time

    # Make a small table with just counts, errors and time for all of the comparison
    # stars.

    comp_fluxes = star_data["date-obs", counts_column_name, error_column_name][good]
    # Convert comp_fluxes to a regular Table, not a QTable, to work around
    # https://github.com/astropy/astropy/issues/10944
    # in which it was reported that QTable columns with units cannot be aggregated.

    comp_fluxes = Table(comp_fluxes)

    # Check whether any of the columns are masked, but with no masked values,
    # and convert to regular column...eventually

    comp_fluxes = comp_fluxes.group_by("date-obs")
    comp_totals = comp_fluxes.groups.aggregate(np.sum)[counts_column_name]
    comp_num_stars = comp_fluxes.groups.aggregate(np.count_nonzero)[counts_column_name]
    comp_errors = comp_fluxes.groups.aggregate(add_in_quadrature)[error_column_name]

    comp_total_vector = np.ones_like(star_data[counts_column_name])
    comp_error_vector = np.ones_like(star_data[error_column_name])

    if len(set(comp_num_stars)) > 1:
        raise RuntimeError("Different number of stars in comparison sets")

    # Calculate relative flux for every star

    # Have to remove the flux of the star if the star is a comparison
    # star.
    # Use the .value below so that we can set the array to 1 and multiply
    # by it without affecting units of the result.
    is_comp = np.zeros_like(star_data[counts_column_name]).value
    is_comp[good] = 1
    flux_offset = -star_data[counts_column_name] * is_comp

    # Convert comp_fluxes back to a QTable and redo groups
    comp_fluxes = QTable(comp_fluxes)
    comp_fluxes = comp_fluxes.group_by("date-obs")
    # This seems a little hacky; there must be a better way
    for date_obs, comp_total, comp_error in zip(
        comp_fluxes.groups.keys, comp_totals, comp_errors, strict=True
    ):
        this_time = star_data["date-obs"] == date_obs[0]
        comp_total_vector[this_time] *= comp_total
        comp_error_vector[this_time] = comp_error * comp_fluxes[error_column_name].unit

    relative_flux = star_data[counts_column_name] / (comp_total_vector + flux_offset)
    relative_flux = relative_flux.flatten()

    rel_flux_error = (
        star_data[counts_column_name]
        / comp_total_vector
        * np.sqrt(
            (star_data[error_column_name] / star_data[counts_column_name]) ** 2
            + (comp_error_vector / comp_total_vector) ** 2
        )
    )

    # Add these columns to table
    if not in_place:
        star_data = star_data.copy()

    star_data["relative_flux"] = relative_flux
    star_data["relative_flux_error"] = rel_flux_error
    star_data["relative_flux_snr"] = relative_flux / rel_flux_error

    # AIJ records the total comparison counts even though that total is used
    # only for the targets, not the comparison.
    star_data["comparison counts"] = comp_total_vector  # + flux_offset
    star_data["comparison error"] = comp_error_vector

    return star_data
